import datetime
import inspect
from collections.abc import Callable
from functools import reduce
from sys import exit

from colored import Back, Fore, Style

from .tcr_dict import clean_dunder_dict
from .tcr_extract_error import extract_error, extract_traceback
from .tcr_getch import getch
from .tcr_iterable import cut_at
from .tcr_print import FMTC, fmt_iterable
from .tcr_terminal import terminal
from .tcr_types import QuotelessString


class CC:
  _ = Style.reset
  LOG = Fore.light_green + Style.bold  # 10
  WARN = Fore.YELLOW + Style.bold
  ERROR = Fore.RED + Style.bold
  DEBUG = Fore.purple_1b + Style.bold  # 129
  CRITICAL = Back.RED + Style.bold
  BANG = Fore.WHITE + Back.RED + Style.bold


class Console:
  """### Provides logging capabilities.

  `console(...)` == `console.debug(...)`
  """

  _last_diff: str | None = None

  @staticmethod
  def _get_timestamp():
    return str(datetime.datetime.now())[:-3].replace('.', ',')

  def log(self, *values, sep=' ', end='', returnonly=False, withprefix=True) -> None | str:
    if not values:
      values = ['']
    out = reduce(lambda x, y: str(x) + sep + str(y), [*values, '']) + end
    if withprefix:
      out = (f'I {self._get_timestamp()} ') + out
    out = f'{CC.LOG}{out}{CC._}'
    if returnonly:
      return out
    print(out)
    return None

  def warn(self, *values, sep=' ', end='', returnonly=False, withprefix=True) -> None | str:
    if not values:
      values = ['']
    out = reduce(lambda x, y: str(x) + sep + str(y), [*values, '']) + end
    if withprefix:
      out = (f'W {self._get_timestamp()} ') + out
    out = f'{CC.WARN}{out}{CC._}'
    if returnonly:
      return out
    print(out)
    return None

  def error(self, *values, sep=' ', end='', returnonly=False, withprefix=True) -> None | str:
    if not values:
      values = ['']
    values = [(extract_error(x) if isinstance(x, Exception) else x) for x in values]
    out = reduce(lambda x, y: str(x) + sep + str(y), [*values, '']) + end
    if withprefix:
      out = (f'E {self._get_timestamp()} ') + out
    out = f'{CC.ERROR}{out}{CC._}'
    if returnonly:
      return out
    print(out)
    return None

  def debug(
    self,
    value: object,
    /,
    *values: object,
    withprefix: bool = True,
    passthrough: bool = True,
    printhook: Callable[[str], None] = print,
    syntax_highlighting=True,
    margin: str = '',
    padding: str = ' ',
    quoteless: bool = False,
    diff: bool = False,
    fmt_iterable: Callable[..., str] = fmt_iterable,
    **kwargs,
  ) -> None | object:
    out = fmt_iterable(*[(x if ((not quoteless) or (not isinstance(x, str))) else QuotelessString(x)) for x in (value, *values)], syntax_highlighting=syntax_highlighting, **kwargs)

    if padding == ' ' and not withprefix:
      padding = ''

    prefix = ''
    if withprefix:
      prefix = f'D {self._get_timestamp()}'

    c_debug = ''
    c_reset = ''
    c_bang = ''
    if syntax_highlighting:
      c_debug = CC.DEBUG
      c_reset = CC._
      c_bang = CC.BANG

    def compose(x) -> str:
      return f'{c_debug}{margin}{prefix}{padding}{x}{c_reset}'

    if diff and not self._last_diff:
      printhook(compose('\n' + '\n'.join(['  ' + x for x in out.split('\n')])))
    else:
      if not diff:
        printhook(compose(out))
      else:
        out_lines = out.split('\n')
        last_lines = self._last_diff.split('\n')

        while len(out_lines) < len(last_lines):
          last_lines.pop(-1)

        while len(last_lines) < len(out_lines):
          last_lines.append('')

        diff_out = '\n'.join([('  ' if line == line_old else f'{c_bang}*{c_reset} ') + line for line, line_old in zip(out_lines, last_lines, strict=True)])

        printhook(compose('\n' + diff_out))

    self._last_diff = out
    return None if not passthrough else value

  def critical(self, *values, sep=' ', end='', returnonly=False, withprefix=True) -> None | str:
    if not values:
      values = ['']
    out = reduce(lambda x, y: str(x) + sep + str(y), [*values, '']) + end
    if withprefix:
      out = (f'C {self._get_timestamp()} ') + out
    out = f'{CC.CRITICAL}{out}{CC._}'
    if returnonly:
      return out
    print(out)
    return None

  def hr(self, **kwargs):
    self.debug(QuotelessString('=' * (terminal.width)), margin='\n', withprefix=False, **kwargs)

  def __call__(self, *args, **kwargs) -> None | str:
    return console.debug(*args, **kwargs)

  def __or__(self, other):
    return self.debug(other)

  def __ror__(self, other):
    return self.debug(other)


console = Console()
"""### Provides logging capabilities.

  `console(...)` == `console.debug(...)`
"""


def _clean_built_in_methods(d: dict) -> dict:
  return {k: v for k, v in d.items() if not isinstance(v, type(abs))}


def start_eval_session(f_backs: int = 2) -> None:
  """Starts an interactive shell. For debugging purposes."""
  try:
    current_frame = inspect.currentframe()
    for _ in range(f_backs):
      current_frame = current_frame.f_back
    desired_locals = current_frame.f_locals
    desired_globals = current_frame.f_globals

    a = ' '
    while a:
      a = input('\n\r>>> ')
      if a:
        try:
          if a.strip().lower() == 'locals':
            console(_clean_built_in_methods(clean_dunder_dict(desired_locals)))
          elif a.strip().lower() == 'globals':
            console(_clean_built_in_methods(clean_dunder_dict(desired_globals)))
          elif a.strip().lower() == 'code':
            lines, start_line = inspect.findsource(current_frame)
            lineno = current_frame.f_lineno

            AROUND = 2
            CUTOFF = 70

            lines = lines[max(0, lineno - (AROUND + 1)) : lineno + AROUND]
            while all(x.startswith(' ') for x in lines):
              lines = [x[1:] for x in lines]
            lines = [f'{FMTC.NUMBER}{lineno-AROUND+i}{Fore.yellow + Style.bold} {">" if i == AROUND else "|"}{FMTC._} {cut_at(x, CUTOFF)}' for i, x in enumerate(lines)]
            lines = ''.join(lines)

            print(lines)
          else:
            console(eval(a, desired_locals, desired_globals))
        except Exception as e:
          console.error(f'\n\n{extract_traceback(e)}\n{extract_error(e)}')
  except (KeyboardInterrupt, EOFError):
    print('\r' + (' ' * terminal.width) + '\r', end='')
    exit()


def breakpoint(*vals, printhook=console, clear=True, ctrlc: Callable[[], None] = start_eval_session) -> None:
  """Stop the program execution until a key is pressed. Optionally pass in things to print."""
  vals and printhook(*vals)
  print(
    (a := f'{Fore.WHITE + Style.BOLD} >>> {Back.RED}BREAKPOINT{Style.RESET} {Fore.WHITE + Style.BOLD}<<<{Style.RESET}'),
    end='\r',
  )

  if getch() == b'\x03':
    if clear:
      print(len(a) * ' ', end='')
    ctrlc()
    return

  if clear:
    print(len(a) * ' ', end='\r')
  else:
    print()



import contextlib
import inspect
import traceback
from types import ModuleType

from .tcr_compare import able
from .tcr_iterable import getattr_queue


def extract_error(e: BaseException, pattern='%s: %s', *, raw=False) -> tuple[str, str] | str:
  if able(issubclass, e, BaseException) and issubclass(e, BaseException):
    tup = (getattr_queue(e, '__name__', '__class__.__name__', default='!UnknownExceptionGroupClass'), '<class>')
    return tup if raw else pattern % tup

  if able(issubclass, e, Exception) and issubclass(e, Exception):
    tup = (getattr_queue(e, '__name__', '__class__.__name__', default='!UnknownExceptionClass'), '<class>')
    return tup if raw else pattern % tup

  if callable(e):
    e = e()

  one = getattr_queue(e, '__name__', '__class__.__name__', default='!UnknownException')
  two = str(e)
  return (one, two) if raw else pattern % (one, two)


def extract_traceback(e: BaseException) -> str:
  traceback_details = traceback.format_tb(e.__traceback__)
  return ''.join(traceback_details)

def module_error_map(module: ModuleType) -> dict[str, BaseException]:
  return {name: obj for name, obj in vars(module).items() if inspect.isclass(obj) and issubclass(obj, BaseException) and not issubclass(obj, Warning)}


def modules_error_map(*modules: ModuleType) -> dict[str, BaseException]:
  return {f'{module.__name__}.{name}'.removeprefix('builtins.'): exc for module in modules for name, exc in module_error_map(module).items()}



from .. import util as __


@__.STRINGIFY
@__.FLATTEN_IF_POSSIBLE
def add(_, *args: str, **ctxs) -> int | float:
  """### Add numbers (supports any number of args).

  Example:
  ```txt
  {add|1|2|3}
  ```

  Result:
  ```txt
  6
  ```
  """
  args = [__.number(x, flatten_to_int_when_possible=True) for x in args]
  return sum(args)


@__.STRINGIFY
@__.REQUIRE_POSITIONAL(1, 0)
@__.FLATTEN_IF_POSSIBLE
def subtract(_, *args: str, **ctxs) -> int | float:
  """### Subtract numbers (supports any number of args).

  Example:
  ```txt
  {add|1|2|3}{#|evaluates to putting a '-' between each number: 1 - 2 - 3}
  ```

  Result:
  ```txt
  -4
  ```
  """
  args = [__.number(x, flatten_to_int_when_possible=True) for x in args]

  if len(args) < 2:
    return args[0]

  return __.functools.reduce(lambda x, y: x - y, args)


@__.STRINGIFY
@__.REQUIRE_POSITIONAL(1, 1)
@__.FLATTEN_IF_POSSIBLE
def multiply(_, *args: str, **ctxs) -> int | float:
  """### Multiply numbers (supports any number of args).

  Example:
  ```txt
  {multiply|2|2|3}
  ```

  Result:
  ```txt
  12
  ```
  """
  args = [__.number(x, flatten_to_int_when_possible=True) for x in args]

  if len(args) < 2:
    return args[0]

  return __.functools.reduce(lambda x, y: x * y, args)


@__.STRINGIFY
@__.REQUIRE_POSITIONAL(1, 1)
@__.FLATTEN_IF_POSSIBLE
def divide(_, *args: str, **ctxs) -> int | float:
  """### Divide numbers (supports any number of args).

  Example:
  ```txt
  {divide|1|2}{#|in case of >2 arguments evaluates to putting a '/' between each number: 1 / 2 / 3}
  ```

  Result:
  ```txt
  0.5
  ```
  """
  args = [__.number(x, flatten_to_int_when_possible=True) for x in args]

  if len(args) < 2:
    return args[0]

  return __.functools.reduce(lambda x, y: x / y, args)


@__.STRINGIFY
@__.REQUIRE_POSITIONAL(1, 1)
@__.FLATTEN_IF_POSSIBLE
def floordivide(_, *args: str, **ctxs) -> int | float:
  """### Floor divide numbers (supports any number of args).

  Example:
  ```txt
  {floordivide|34|3}{#|in case of >2 arguments evaluates to putting a '//' between each number: 1 // 2 // 3}
  ```

  Result:
  ```txt
  11
  ```
  """
  args = [__.number(x, flatten_to_int_when_possible=True) for x in args]

  if len(args) < 2:
    return int(args[0])

  return __.functools.reduce(lambda x, y: x // y, args)


@__.STRINGIFY
@__.REQUIRE_POSITIONAL(1, None)
@__.FLATTEN_IF_POSSIBLE
def power(_, base: str, exponent: str = '2', *args: str, **ctxs) -> int | float:
  args = [__.number(x, flatten_to_int_when_possible=True) for x in (base, exponent, *args)]

  return __.functools.reduce(lambda x, y: x**y, args)


@__.STRINGIFY
@__.REQUIRE_POSITIONAL(1, None)
@__.FLATTEN_IF_POSSIBLE
def modulo(_, *args: str, **ctxs) -> int | float:
  """### Modulo numbers (supports any number of args).

  Example:
  ```txt
  {modulo|34|3}{#|in case of >2 arguments evaluates to putting a '%' between each number: 1 % 2 % 3}
  ```

  Result:
  ```txt
  1
  ```
  """
  args = [__.number(x, flatten_to_int_when_possible=True) for x in args]

  if len(args) < 2:
    return args[0]

  return __.functools.reduce(lambda x, y: x % y, args)


@__.STRINGIFY
@__.REQUIRE_POSITIONAL(1, None)
def round_(_, n: str, *args: str, **ctxs) -> int | float:
  """### Round numbers using python's round function. (This DISCARDS any extra arguments!).

  This does not just round halves up, see: https://en.wikipedia.org/wiki/Rounding#Rounding_half_to_even

  Example:
  ```txt
  {round|1.2}
  {round|1.5}
  {round|2.5}
  {round|3.5}
  {round|4.5}
  ```

  Result:
  ```txt
  1
  2
  2
  4
  4
  ```
  """
  return round(__.number(n))


@__.STRINGIFY
@__.REQUIRE_POSITIONAL(1, None)
def floor(_, n: str, *args: str, **ctxs) -> int | float:
  """### Round down. (This DISCARDS any extra arguments!).

  Example:
  ```txt
  {floor|6.1}
  {floor|6.9}
  ```

  Result:
  ```txt
  6
  6
  ```
  """
  return __.math.floor(__.number(n))


@__.STRINGIFY
@__.REQUIRE_POSITIONAL(1, None)
def ceil(_, n: str, *args: str, **ctxs) -> int | float:
  """### Round up. (This DISCARDS any extra arguments!).

  Example:
  ```txt
  {ceil|9.1}
  {ceil|9.9}
  ```

  Result:
  ```txt
  10
  10
  ```
  """
  return __.math.ceil(__.number(n))



from types import UnionType
from typing import Any, Self, TypeVar, Union

from .tcr_class import Singleton


def instance(func):
  return func()


T = TypeVar('T')


class _NullLike:
  """Base class for null-like objects (Null, Undefined).

  Can be bitwise-ored with any other value to yield that other value. (If both values in bitwise or are null-like, return the directionally right one)
  """

  def __str__(self):
    return self.__repr__()

  def __repr__(self):
    return self.__class__.__name__

  def __bool__(self):
    return False

  def __or__(self, __value: T) -> T:
    return __value

  def __ror__(self, __value: T) -> T:
    return __value


@instance
class Null(Singleton, _NullLike):
  """Value is explicitly declared as absent."""


@instance
class Undefined(Singleton, _NullLike):
  """Value is missing."""


class UniqueDefault:
  """Used as a default parameter for the parameter named default (yeah i know that's uh... let's just see the example).

  Example:
  ```py
  from tcrutils import UniqueDefault as RaiseError

  def get_something(..., *, default: Any | RaiseError = RaiseError):
  ... # Function implementation

  # Nothing was found, falling to raising error or returning default value if such was specified.
  if default is not RaiseError:
    return default
  raise KeyError("Unable to find something.")
  """



from collections.abc import Callable
from typing import Any, NoReturn


def void(*args: Any, **kwargs: Any) -> None:
  """Synchronous voider: take any arguments and do nothing, useful in functions that require a callback when nothing is needed to be done."""


async def avoid(*args: Any, **kwargs: Any) -> None:
  """Asynchronous voider: take any arguments and do nothing, useful in functions that require a callback when nothing is needed to be done."""


def raiser(e: Exception) -> Callable[..., NoReturn]:
  """With decorator-like structure return a synchronous callable which raises specified exception on call, no matter what (with *args, **kwargs which are ignored)."""

  def inner_raiser(*args, **kwargs):
    raise e

  return inner_raiser


async def araiser(e: Exception) -> Callable[..., NoReturn]:
  """With decorator-like structure return an asynchronous callable which raises specified exception on call, no matter what (with *args, **kwargs which are ignored)."""

  async def inner_raiser(*args, **kwargs):
    raise e

  return inner_raiser



from pydantic import BaseModel as _BaseModel


class _BM(_BaseModel):
  def __repr__(self):
    # Get the class name
    class_name = self.__class__.__name__
    # Get a string of all field names and values, with space separation
    fields_str = " ".join(f"{key}={value!r}" for key, value in self.__dict__.items())
    # Format the string as required
    return f"{class_name}({fields_str})"

# Example usage
class ExampleModel(_BM):
  a: int
  b: int
  c: int




try:
  from . import placeholder_sets as placeholder_set
  from . import placeholders as placeholder
  from .core import (
    DynamicResponseBuilder,
    DynamicResponseError,
    DynamicResponseMissingPlaceholderError,
    DynamicResponsePlaceholderInvalidReturnError,
    DynamicResponseRecursionError,
    DynamicResponseSyntaxError,
  )
except ImportError: ...



import base64


class b64:
  def __init__(self) -> None:
    raise RuntimeError('This class should not be instantiated.')

  @staticmethod
  def encode(text: str, encoding: str = 'utf-8') -> str:
    return base64.b64encode(text.encode(encoding=encoding)).decode(encoding=encoding)

  @staticmethod
  def decode(text: str, encoding: str = 'utf-8') -> str:
    return base64.b64decode(text.encode(encoding=encoding)).decode(encoding=encoding)



from collections.abc import Callable, Mapping
from typing import Any

from ..discord.tcrd_types import HikariDictMessage
from ..src.tcr_console import breakpoint as bp
from ..src.tcr_dict import merge_dicts
from ..src.tcr_run import run_sac
from .error import *
from .util import _TT, _Token, build_placeholder_rich_return_value

### Lexer


class _DRLexer:
  text: str
  parens: tuple[str, str]
  pos: int
  tokens: list[_Token]

  def __init__(self, text: str, parens: tuple[str, str]) -> None:
    self.text = text
    self.parens = parens
    self.pos = -1  # Advancing first, then checking, so it's 1 less
    self.tokens = []

  def _advance(self, amount: int = 1) -> None:
    self.pos += amount

  def starts_with(self, s: str) -> bool:
    return self.text[self.pos : self.pos + len(s)] == s

  def fetch_text(self) -> str:
    if self.pos >= len(self.text):
      return ''
    s = ''
    while self.pos < len(self.text) and not self.starts_with(self.parens[0]) and not self.starts_with(self.parens[1]):
      s += self.text[self.pos]
      self._advance()

    self._advance(-1)
    return s

  def __call__(self) -> list[_Token]:
    while self.pos < len(self.text):
      self._advance()
      if self.starts_with(self.parens[0]):
        self.tokens.append(_Token(_TT.PAREN_OPEN))
        self._advance(len(self.parens[0]) - 1)
      elif self.starts_with(self.parens[1]):
        self.tokens.append(_Token(_TT.PAREN_CLOSE))
        self._advance(len(self.parens[1]) - 1)
      else:
        self.tokens.append(_Token(_TT.TEXT, self.fetch_text()))
    return self.tokens[:-1]


### Parser


class _DRParser:
  contexts: Mapping[str, Any]
  placeholders: dict[str | None, Callable[..., str]]
  tokens: list[_Token]
  error_on_missing_placeholder: bool

  def __init__(
    self,
    contexts: Mapping[str, Any],
    placeholders: dict[str, Callable[..., str]],
    splitter: str,
    tokens: tuple[_Token],
    error_on_missing_placeholder: bool,
    error_on_invalid_placeholder_return: bool,
    parens: tuple[str, str],
  ) -> None:
    self.contexts = {**contexts, '__splitter': splitter, '__parens': parens, '__vars': {}}
    self.placeholders = placeholders
    self.tokens = list(tokens)
    self.error_on_missing_placeholder = error_on_missing_placeholder
    self.error_on_invalid_placeholder_return = error_on_invalid_placeholder_return

  def __tcr_fmt__(self, fmt_iterable, **kwargs):
    return self.__repr__()

  def check_integrity(self) -> None:
    parens_counter = 0

    for arg in self.tokens:
      if arg.type == _TT.PAREN_OPEN:
        parens_counter += 1
      elif arg.type == _TT.PAREN_CLOSE:
        parens_counter -= 1

      if parens_counter < 0:
        raise DynamicResponseSyntaxError(f'Mismatched parenthesis (Closed an unopened section.)')
    if parens_counter != 0:
      raise DynamicResponseSyntaxError(f'Mismatched parenthesis (Open section left unclosed.)')

  async def process_single_placeholder(self, tokens: list[_Token]) -> list[_Token]:
    text: str = ''.join(x.value for x in tokens)

    args = text.split(self.contexts['__splitter'])

    if args[0] in self.contexts['__vars']:
      return [_Token(_TT.TEXT, self.contexts['__vars'][args[0]])]
    if self.error_on_missing_placeholder and args[0] not in self.placeholders:
      raise DynamicResponseMissingPlaceholderError(f'Placeholder {args[0]!r} not found.')
    if args[0] not in self.placeholders:
      return [_Token(_TT.TEXT, f'{self.contexts["__parens"][0]}{text}{self.contexts["__parens"][1]}')]
    ret = await run_sac(self.placeholders[args[0]], *args, **self.contexts)
    if ret is None:
      ret = ''

    if type(ret) is not str:  # noqa: E721
      if isinstance(ret, list | tuple):
        if all(isinstance(x, _Token) for x in ret):
          return list(ret)
        else:
          raise DynamicResponseSyntaxError(f'Placeholder {args[0]!r} returned a list but contains non-token elements.')
    else:
      return [_Token(_TT.TEXT, ret)]

    if self.error_on_invalid_placeholder_return:
      raise DynamicResponsePlaceholderInvalidReturnError(f'Placeholder {args[0]!r} did not return a string nor list of tokens. (ret={ret!r})')
    return [_Token(_TT.TEXT, str(ret))]

  async def process_placeholders(self, tokens: list[_Token]) -> list[_Token]:
    first_open_index = None
    first_close_index = None

    for token in tokens:
      if token.type == _TT.PAREN_OPEN:
        first_open_index = tokens.index(token)
      if token.type == _TT.PAREN_CLOSE:
        first_close_index = tokens.index(token)
        break

    if first_open_index is None and first_close_index is None:
      return tokens
    elif first_open_index is None or first_close_index is None:
      raise RuntimeError('Internal error how the fuck are parenthesis mismatched again')

    for token in tokens[first_open_index + 1 : first_close_index]:
      if token.type != _TT.TEXT:
        raise RuntimeError('Shit is broken again (internal error) report this fucking piece of shit not fucking working FUUUUCK!!!!!!!!')

    placeholder_text_tokens = tokens[first_open_index + 1 : first_close_index]

    try:
      # print(f"[DEBUG] Processing {'|'.join(x.value for x in placeholder_text_tokens)}")
      processed_text_tokens = await self.process_single_placeholder(placeholder_text_tokens)
    # except BaseException:
    #   ...
    except DynamicResponseError:
      raise
    except RecursionError as e:
      raise DynamicResponseRecursionError(f'{e}') from e
    except TypeError as e:
      if 'missing' in str(e) and 'required positional argument' in str(e):
        raise DynamicResponsePlaceholderTooFewArgumentsError(f'Placeholder {placeholder_text_tokens} requires more arguments than were provided.') from e
      raise DynamicResponsePythonErrorInPlaceholerError(f'An error occured within a placeholder {placeholder_text_tokens}.') from e
    except Exception as e:
      raise DynamicResponsePythonErrorInPlaceholerError(f'An error occured within a placeholder {placeholder_text_tokens}.') from e

    del tokens[first_open_index : first_close_index + 1]
    for tkn in processed_text_tokens[::-1]:
      tokens.insert(first_open_index, tkn)

    return await self.process_placeholders(tokens)

  async def __call__(self) -> tuple[str, dict]:
    self.check_integrity()

    processed = await self.process_placeholders(self.tokens[:])
    return ''.join(x.value for x in processed), self.contexts


### Dynamic Response Builder

RESP_CONTEXTS = (
  'content',
  'attachments',
  'components',
  'embed',
  'stickers',
  'tts',
  'reply',
  'reply_must_exist',
  'mentions_everyone',
  'mentions_reply',
  'user_mentions',
  'role_mentions',
  'flags',
)


class DynamicResponseResult(str):
  contexts: dict[str, Any]

  def __new__(cls, string, contexts):
    instance = super().__new__(cls, string)
    instance.contexts = contexts
    return instance

  @property
  def resp(self) -> HikariDictMessage:
    return merge_dicts({'content': self}, {x: y for x, y in self.contexts.items() if x in RESP_CONTEXTS})


class DynamicResponseBuilder:
  placeholders: dict[str, Callable]
  parens: tuple[str, str]
  splitter: str
  instance_contexts: dict[str, Any]

  def __init__(
    self,
    *placeholders: dict[str, Callable] | None,
    parens: tuple[str, str] = ('{', '}'),
    splitter: str = '|',
    error_on_missing_placeholder: bool = True,
    error_on_invalid_placeholder_return: bool = True,
    instance_contexts: dict[str, Any] | None = None,
    context_constructors: dict[str, Callable[[], Any]] = {},  # noqa: B006
  ) -> None:
    if not all(isinstance(x, dict) for x in placeholders):
      raise TypeError('placeholders must be a dict or list of dicts')
    placeholders = merge_dicts(*placeholders)

    self.placeholders = placeholders
    self.parens = parens
    self.splitter = splitter
    self.error_on_missing_placeholder = error_on_missing_placeholder
    self.error_on_invalid_placeholder_return = error_on_invalid_placeholder_return
    self.instance_contexts = instance_contexts or {}
    self.context_constructors = context_constructors

  async def __call__(self, text: str, **contexts: Any) -> DynamicResponseResult:
    tokens = _DRLexer(text, parens=self.parens)()
    parsed, contexts = await _DRParser(
      contexts={**self.instance_contexts, **{x: y() for x, y in self.context_constructors.items()}, **contexts},
      placeholders=self.placeholders,
      splitter=self.splitter,
      tokens=tokens,
      error_on_missing_placeholder=self.error_on_missing_placeholder,
      error_on_invalid_placeholder_return=self.error_on_invalid_placeholder_return,
      parens=self.parens,
    )()

    return DynamicResponseResult(parsed, contexts)  # noqa: RET504 , RUF100

  def add_placeholder(self, name: str, func: Callable) -> None:
    self.placeholders[name] = func



import pathlib as p
import shelve
import shutil
import string
from collections.abc import Callable, Iterator
from typing import Any, Self

ALLOWED_CHARACTERS = string.ascii_letters + string.digits + "!&#'^~$,.%`{}[]();@_-+="  # Allowed characters in DB ID

DISALLOWED_SEQUENCES = (
  'PRN',
  'CON',
  'AUX',
  'NUL',
  'COM1',
  'COM2',
  'COM3',
  'COM4',
  'COM5',
  'COM6',
  'COM7',
  'COM8',
  'COM9',
  'COM0',
  'LPT1',
  'LPT2',
  'LPT3',
  'LPT4',
  'LPT5',
  'LPT6',
  'LPT7',
  'LPT8',
  'LPT9',
  'LPT0',
)


def c(a):
  return f'\x1b[{a}m'


A = f"""
{c("208")}class {c("39")}DB{c("7")}({c}{c(39)}tcr.ShelveDB{c("7")}){c}{c(231)}:
  {c(7)}directory {c(208)}= {c(46)}"/VALID/path/to/db/dir/" {c(8)}# Or pathlib.Path object{c}
"""[1:-1]


class ShelveDB(dict):
  """### Wrapper for shelve module databases.

  ```py
  class DB(tcr.ShelveDB):
    # REQUIRED:
    directory = "/VALID/path/to/db/dir/" # Or pathlib.Path object
    # OPTIONAL:
    defaults = {"r": list}
  ```
  ### Defaults:
  If key "r" is accessed but it's not found, return `list()` (that'd be a `[]`, not the `list` type itself of course!) and set the value at that key to it, so the next time it's looked up a value will be present, unless deleted or overwritten in the meantime.
  Values must be any zero-argument callables. (functions, classes/types, etc.).
  This is to avoid mutability problems, for each key a new instance is created.

  ## Raises:
    - ValueError: provided alnum_id is invalid.
    - RuntimeError: You did not or incorrectly set up the class declaration (shown above).
    - NotADirectoryError: The directory path you provided points to a file, not a directory or nothing.
  """

  directory: str | p.Path = None
  __directory: p.Path
  s: shelve.Shelf
  defaults: dict[Any, Callable[[], Any]] = None

  def __init__(self, alnum_id: str | int) -> None:
    if isinstance(self.directory, str | p.Path):
      if isinstance(self.directory, str):
        self.directory = p.Path(self.directory)
      self.__directory = self.directory
    else:
      raise RuntimeError('Set up the db as follows:\n' + A)  # noqa: TRY004

    if self.__directory.is_file():
      raise NotADirectoryError('Provide a path to a directory or a nonexistent path ' "(a directory will be created if it doesn't exist).")

    self.__directory.mkdir(exist_ok=True, parents=True)

    if self.defaults is None:
      self.defaults = {}

    if not all(callable(v) for v in self.defaults.values()):
      raise RuntimeError('All values of defaults must be callable.')

    if not all(char in ALLOWED_CHARACTERS for char in str(alnum_id)):
      raise ValueError(f'alnum_id must be any of those characters: {ALLOWED_CHARACTERS}')

    if str(alnum_id) in DISALLOWED_SEQUENCES:
      raise ValueError(f'alnum_id must not be any of theese: {", ".join(DISALLOWED_SEQUENCES)}')

    self.alnum_id = str(alnum_id)
    alnum_dirpath = self.__directory / self.alnum_id
    alnum_dirpath.mkdir(exist_ok=True)
    self.s = shelve.open(alnum_dirpath / self.alnum_id)

    super().__init__(self.s)

  def __getitem__(self, __key: Any) -> Any:
    try:
      return super().__getitem__(__key)
    except KeyError:
      if __key in self.defaults:
        val = self.defaults[__key]()
        self[__key] = val
        return val
      else:
        raise

  def __setitem__(self, __key: Any, __value: Any) -> None:
    self.s[__key] = __value
    self.s.sync()
    return super().__setitem__(__key, __value)

  def __delitem__(self, __key: Any) -> None:
    del self.s[__key]
    self.s.sync()
    return super().__delitem__(__key)

  def __del__(self) -> None:
    try:
      self.s.close()
    except:  # noqa: E722
      return

  @property
  def d(self):
    """Convert the underlying shelf into a dictionary (make a copy as dict)."""
    return dict(self.s)  # .copy()

  def clear(self) -> None:
    """Clear both shelf and the underlying dictionary."""
    self.s.clear()
    self.s.sync()
    super().clear()

  def update(self, **kwargs: Any) -> None:
    """Update both the shelf and the underlying dictionary."""
    self.s.update(**kwargs)
    self.s.sync()
    super().update(**kwargs)

  def copy(self) -> dict:
    """Relinquishes the database connection and returns the underlying's dict copy."""
    return super().copy()

  def pop(self, *args, **kwargs) -> Any:
    super().pop(*args, **kwargs)
    res = self.s.pop(*args, **kwargs)
    self.s.sync()
    return res

  def popitem(self) -> tuple[Any, Any]:
    res = self.s.popitem()
    super().pop(res[0])
    self.s.sync()
    return res

  def setdefault(self, key, default=None):
    super().setdefault(key, default)
    res = self.s.setdefault(key, default)
    self.s.sync()
    return res

  def get_directory(self) -> p.Path:
    return self.__directory

  def get_path(self) -> p.Path:
    return self.get_directory() / self.alnum_id

  def drop_db(self) -> None:
    """Close the shelf and delete the underlying system directory of this database instance."""
    self.s.close()
    shutil.rmtree(self.__directory / self.alnum_id)

  @classmethod
  def iter_all(cls) -> Iterator[tuple[str, Self]]:
    for path in p.Path(cls.directory).iterdir():
      yield path.name, cls(path.name)

  @classmethod
  def iter_all_paths(cls) -> Iterator[p.Path]:
    yield from p.Path(cls.directory).iterdir()

  @classmethod
  def iter_all_path_names(cls) -> Iterator[str]:
    for path in p.Path(cls.directory).iterdir():
      yield path.name

  @classmethod
  def iter_all_shelves(cls) -> Iterator[Self]:
    for path in p.Path(cls.directory).iterdir():
      yield cls(path.name)

  @classmethod
  def exists(cls, alnum_id: str | int) -> bool:
    """Check if the given `alnum_id` is already in use (if a database directory exists and is a directory and is not empty)."""
    folder = p.Path(cls.directory) / str(alnum_id)
    return folder.is_dir() and (len(list(folder.iterdir())) > 0)  # Counting empty directories as non existent because theres no data



"""Contains joke functions or the ones that will never be used in a serious situations/prod (This entire package sucks so much it should never be used in prod but uhhhhh.....)."""

from collections.abc import Callable
from typing import Any, Literal, Self

from ..src import tcr_cloud_imports as cloud_imports
from .tcr_decorator import instance


def fizzbuzz(n: int) -> str: ...


fizzbuzz: Callable[[int], str] = lambda n: 'Fizz' * (n % 3 == 0) + 'Buzz' * (n % 5 == 0) or str(n)  # fmt: skip  # noqa: F811


def oddeven(n: int | str) -> Literal['odd', 'even']:
  """### Outputs a string `'odd'` or `'even'` based on the supplied int `n`.

  ```py
  >>> from tcrutils.other import oddeven
  >>> oddeven(1)
  'odd'
  >>> oddeven(2)
  'even'
  """
  n = int(n)
  return 'eovdedn'[n % 2 :: 2]


def christmas_tree(*, height=10, symbol='*'):
  """Generate a christmas tree for printing in a console.

  Height is the number of lines the tree will have.\\
  Symbol is any 1-2 long string for example 'C#' or '*'.
  """
  if len(symbol) == 1:
    symbol = 2 * symbol

  def transform(symbols: str):
    return symbols.center(height * 2, ' ')

  return '\n'.join([transform(symbol * x) for x in range(height + 1) if x])


def _cint_decorate_int_method(func):
  def wrapper(*args, **kwargs):
    result = func(*args, **kwargs)
    if isinstance(result, int):
      return CInt(result)
    else:
      return result

  return wrapper


class CInt:
  def __init__(self, value):
    self.value = value
    self.pos_flag = False
    self.neg_flag = False

  def __pos__(self):
    if self.pos_flag:
      self.value += 1
    self.pos_flag = not self.pos_flag
    return self

  def __neg__(self):
    if self.neg_flag:
      self.value -= 1
    self.neg_flag = not self.neg_flag
    return self

  def __eq__(self, other: Self | int) -> bool:
    if isinstance(other, CInt):
      return self.value == other.value
    else:
      return self.value == other

  def __int__(self):
    return self.value

  def __str__(self):
    return str(self.value)

  def __repr__(self) -> str:
    return self.__str__()

  def __tcr_fmt__(self, fmt_iterable, **kwargs):
    if self.neg_flag or self.pos_flag:
      return fmt_iterable(complex(int(self), 0.5 * (self.pos_flag - self.neg_flag)), **kwargs)
    else:
      return fmt_iterable(int(self), **kwargs)

  def __getattr__(self, name):
    if hasattr(self.value, name) and callable(getattr(self.value, name)):
      method = getattr(self.value, name)
      return _cint_decorate_int_method(method)
    else:
      raise AttributeError(f"'CInt' object has no attribute '{name}'")


class Private:
  """Makes every attribute private, like in other languages that support private/public attribute modifiers.

  Unfortunately i haven't gotten to writing a Public version of it so you can only private every single field :3

  Usage:
  ```py

  class PrivateString(Private, str): ...

  s = PrivateString('abc')

  print(s) # -> abc
  print(s.upper) # -> AttributeError: 'PrivateString' object has no attribute 'upper'
  print(s.__getattribute__) # -> AttributeError: 'PrivateString' object has no attribute '__getattribute__'

  ```
  """

  def __getattribute__(self, name):
    class_name = object.__getattribute__(self, '__class__').__name__
    raise AttributeError(f'{class_name!r} object has no attribute {name!r}')


@instance
class gmail:
  def __init__(self) -> None:
    self.tlds = []

  def __getattr__(self, name):
    self.tlds.append(f'{name}')
    return self

  def __rmatmul__(self, other):
    class _Email:
      def __init__(self, email: str) -> None:
        self.email = email

      def __str__(self) -> str:
        return self.email

      def __repr__(self) -> str:
        return repr(self.__str__())

      def send(self, title: str, body: str) -> None:
        print('Sent email to ' + self.email + ' with title: ' + repr(title) + ', and body: ' + repr(body))

    try:
      return _Email(f'{other!s}@{self.__class__.__name__}.{".".join(self.tlds)}')
    finally:
      self.tlds.clear()



"""TCRUtils Discord. Contains various features related to bot development on discord and other stuff maybe."""

from .tcrd_alias import get_guild_count
from .tcrd_constants import DISCORD_EPOCH
from .tcrd_dpy import escape_markdown, escape_mentions, remove_markdown
from .tcrd_limits import DiscordLimits
from .tcrd_snowflake import is_snowflake
from .tcrd_string import IFYs, get_token

try:
  from . import tcrd_types as types
  from .tcrd_embeds import embed, modal
  from .tcrd_permissions import PERMISSIONS_DICT, Permission, permissions
  from .tcrd_permissions import Permission as DiscordPermission
  from .tcrd_permissions import permissions as discord_permissions
  from .tcrd_shorts import confirm
except ImportError: ...



import re as regex
from collections.abc import Callable, Iterable, Iterator, MutableSequence
from random import shuffle
from typing import Any, Literal, TypeVar

from .tcr_compare import able
from .tcr_null import UniqueDefault as RaiseError
from .tcr_regex import RegexPreset

T = TypeVar('T')
Ts = TypeVar('Ts')


def batched(
  it: Iterable[T],
  n: int,
  *,
  back_to_front: bool = False,
) -> list[Iterable[T]]:
  """### Poor man's `itertools.batched()` (on Py3.11).

  itertools.batched exists but I left it here because:
  - I still need to support py3.11
  - I don't think itertools.batched() has an option to easily back_to_front the results.

  Return a list of splits of the original `it` Iterable that was passed in, split every `n` items. Last group may be smaller than `n` items if the source was exhausted
  `back_to_front`: makes the items pile up in the back and now the first item is the one to have less than n items if there's not enough to pack it with. For example:
  - `batched("1234567890", n=3)                    ` returns `["123", "456", "789", "0"]`
  - `batched("1234567890", n=3, back_to_front=True)` returns `["1", "234", "567", "890"]`
  """
  if back_to_front:
    return [x[::-1] for x in batched(it[::-1], n=n, back_to_front=False)[::-1]]
  return [it[i : i + n] for i in range(0, len(it), n)]


def cut_at(
  it: str | Iterable,
  n: int,
  end: str | Iterable = '...',
  *,
  filter_links: Literal[False] | str | Callable = False,
  shrink_links_visually_if_fits: bool = False,
) -> str:
  """### Return a cut off part of the provided `it` iterable.

  with the `end` added at the end such that the return value is exactly `n` in length.\\
  Mainly strings but also supports other iterables\\
  Set `end` to `''`/`[]` to disable it.\\

  Use `filter_links` to shorten links to this string, only applicable if the iterable provided is a string and if the iterable exceeded `n`, else nothing will be affected
  - `$1` - Entire link
  - `$2` - http/https (will keep the 's' if present, that's its point)
  - `$3` - any `subdomains.` (if present), `.domains.` or `.tld`s
  - `$4` - `.domain` (second to last (possibly after . and) before the last .)
  - `$5` - `.tld` (in between the last . and / or end of match)
  - `$6` - part of the link after `/` (not including it, may be '')

  https://regex101.com/r/46LZtS/1

  `shrink_links_visually_if_fits` replaces links with their markdown counterparts where the visible text is http(s):/subdomain.domain.tld/ (no further part of the links is visible, but the entire link is contained)
  """
  if len(it) <= n:
    if shrink_links_visually_if_fits and len(a := regex.sub(RegexPreset.URL, r'[\2:/\3/](<\1>)', it)) <= n:  # :/ instead of :// because discord bruh
      return a
    return it
  if filter_links is not False and isinstance(it, str):
    if not isinstance(filter_links, str | Callable):
      msg = f'Set filter_links to a str of the replacement value for links or Callable, not {filter_links!r}'
      raise ValueError(msg)
    it = regex.sub(RegexPreset.URL, filter_links, it)
    if len(it) <= n:
      return it
  if n > len(end):
    return it[: (n - len(end))] + end
  if n > 0:
    return end[:n]
  return ''


def bogo_sort(arr: list[T]) -> list[T]:
  if 0 <= len(arr) <= 1:
    return arr
  while True:
    shuffle(arr)
    for i in range(len(arr) - 1):
      if arr[i] > arr[i + 1]:
        break
    else:
      return arr


def stalin_sort(arr: Iterable[T]) -> list[T]:
  if len(arr) <= 1:
    return arr

  sorted_arr = [arr[0]]

  for i in range(1, len(arr)):
    if arr[i] >= sorted_arr[-1]:
      sorted_arr.append(arr[i])

  return sorted_arr


def shuffled(arr: MutableSequence[T]) -> MutableSequence[T]:
  shuffle(arr)
  return arr


def limited_iterable(it: Iterable[T] | Iterator[T], limit: int) -> tuple[list[T], int]:
  """## Return tuple[list, int] of (limited_iterable, overflow_from_limit).

  ### This is different from it[:limit] because it works with iterators

  Examples:
  ```py
  >>> limited_iterable([10, 20, 30], 10)
  ([10, 20, 30], 0) # Nothing is cut off
  >>> limited_iterable([10, 20, 30], 3)
  ([10, 20, 30], 0)
  >>> limited_iterable([10, 20, 30], 2)
  ([10, 20], 1)
  >>> limited_iterable([10, 20, 30], 1)
  ([10], 2)
  >>> limited_iterable([10, 20, 30], 0)
  ([], 3)
  >>> limited_iterable([10, 20, 30], -1)
  ([], 4) # Overflow counter counts the negative limit even though it makes no sense, for predictability reasons
  ```
  ---

  ### When there's no len method on the passed in iterable (eg. generator) and the iterable takes longer than the limit the overflow returned is -1**

  ```py
  >>> limited_iterable(range(10), 7)
  ([0, 1, 2, 3, 4, 5, 6], 3) # range by itself has len support even though it's a generator due to comparing boundaries
  >>> limited_iterable(iter(range(10)), 7)
  ([0, 1, 2, 3, 4, 5, 6], -1) # when wrapped into an iter() the length can't be determined within the limit
  >>> limited_iterable(iter(range(5)), 7)
  ([0, 1, 2, 3, 4], 0) # 0 was returned even though it's a non-len object because the iterator was exhausted and the end of it being within limit could be determined.
  ```
  """
  if limit <= 0:
    return ([], len(it) - limit if able(len, it) else -1)
  list_ = []
  breaknext = False
  for i, item in enumerate(it):
    if breaknext:
      return breaknext
    list_.append(item)
    if i + 2 > limit:
      overflow = len(it) - limit if able(len, it) else -1
      breaknext = (list_, overflow)
  return list_, 0


def hasmanyattrs(obj, attr: str, *attrs: str) -> bool:
  attrs = (attr, *attrs)

  for attr in attrs:
    if not hasattr(obj, attr):
      return False

    obj = getattr(obj, attr)

  return True


def getmanyattrs(obj, attr: str, *attrs: str) -> Any:
  attrs = (attr, *attrs)

  for attr in attrs:
    obj = getattr(obj, attr)

  return obj


def getattr_queue(
  obj: object,
  /,
  *queue: str,
  return_as_str: bool = False,
  default: Any | RaiseError = RaiseError,
):
  """## Try to access `obj`'s attrs in order.

  ```py
  queue = ('__qualname__', '__name__')

  getattr(obj, '__qualname__') # return this if hasattr(obj, '__qualname__') else go to the next line
  getattr(obj, '__name__')     # return this if hasattr(obj, '__name__') else go to the next line
  if default is not RaiseError # RaiseError is the default value for default (if nothing is passed in)
    return default
  raise KeyError("Unable to find any attrs from queue in obj.") # Nothing could be found, error raised.
  ```

  In other words: try to get name #1, if not found, get name #2, if not found get #3, etc. etc. until the queue is exhausted.
  """
  if not queue:
    raise ValueError('queue cannot be empty.')
  for name in queue:
    if hasmanyattrs(obj, *name.split('.')):
      if return_as_str:
        return name
      else:
        return getmanyattrs(obj, *name.split('.'))
  if default is not RaiseError:
    return default
  raise KeyError('Unable to find any attrs from queue in obj.')


def Or(arg: T, *args: Ts, none: Any = None) -> T | Ts:
  """### Returns the first element of the tuple (arg, *args) that does not equal the supplied (`none`) variable, by default None.

  This is different from or chaining because this does not check for truthy values but rather non-none values.\\
  Keep in mind, the passed in none may be unrelateed to built-in None however that is its default.
  """
  args = (arg, *args)

  for a in args:
    if a != none:
      return a

  return args[-1]



"""### This code was not written by me, however it's licensed with the MIT License which allows for free copy and sublicensing of the licensed material.

Also i ain't depending on a whole ass package just for markdown escaping.

This code was copied from the [discord.py](https://github.com/Rapptz/discord.py) python package.

Copy of the original license (applies only to this file):

```txt
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
```

Small modifications were made to this code
"""

import re

_MARKDOWN_ESCAPE_SUBREGEX = '|'.join(r'\{0}(?=([\s\S]*((?<!\{0})\{0})))'.format(c) for c in ('*', '`', '_', '~', '|'))  # noqa: UP032
_MARKDOWN_ESCAPE_COMMON = r'^>(?:>>)?\s|\[.+\]\(.+\)'
_MARKDOWN_ESCAPE_REGEX = re.compile(rf'(?P<markdown>{_MARKDOWN_ESCAPE_SUBREGEX}|{_MARKDOWN_ESCAPE_COMMON})', re.MULTILINE)
_URL_REGEX = r'(?P<url><[^: >]+:\/[^ >]+>|(?:https?|steam):\/\/[^\s<]+[^<.,:;\"\'\]\s])'
_MARKDOWN_STOCK_REGEX = rf'(?P<markdown>[_\\~|\*`#-]|{_MARKDOWN_ESCAPE_COMMON})'


def remove_markdown(text: str, *, ignore_links: bool = True) -> str:
  """A helper function that removes markdown characters.

  .. versionadded:: 1.7

  .. note::
          This function is not markdown aware and may remove meaning from the original text. For example,
          if the input contains ``10 * 5`` then it will be converted into ``10  5``.

  Parameters
  -----------
  text: :class:`str`
      The text to remove markdown from.
  ignore_links: :class:`bool`
      Whether to leave links alone when removing markdown. For example,
      if a URL in the text contains characters such as ``_`` then it will
      be left alone. Defaults to ``True``.

  Returns
  --------
  :class:`str`
      The text with the markdown special characters removed.
  """

  def replacement(match):
    groupdict = match.groupdict()
    return groupdict.get('url', '')

  regex = _MARKDOWN_STOCK_REGEX
  if ignore_links:
    regex = f'(?:{_URL_REGEX}|{regex})'
  return re.sub(regex, replacement, text, count=0, flags=re.MULTILINE)


def escape_markdown(text: str, *, as_needed: bool = False, ignore_links: bool = True) -> str:
  r"""A helper function that escapes Discord's markdown.

  Parameters
  -----------
  text: :class:`str`
      The text to escape markdown from.
  as_needed: :class:`bool`
      Whether to escape the markdown characters as needed. This
      means that it does not escape extraneous characters if it's
      not necessary, e.g. ``**hello**`` is escaped into ``\*\*hello**``
      instead of ``\*\*hello\*\*``. Note however that this can open
      you up to some clever syntax abuse. Defaults to ``False``.
  ignore_links: :class:`bool`
      Whether to leave links alone when escaping markdown. For example,
      if a URL in the text contains characters such as ``_`` then it will
      be left alone. This option is not supported with ``as_needed``.
      Defaults to ``True``.

  Returns
  --------
  :class:`str`
      The text with the markdown special characters escaped with a slash.
  """

  if not as_needed:

    def replacement(match):
      groupdict = match.groupdict()
      is_url = groupdict.get('url')
      if is_url:
        return is_url
      return '\\' + groupdict['markdown']

    regex = _MARKDOWN_STOCK_REGEX
    if ignore_links:
      regex = f'(?:{_URL_REGEX}|{regex})'
    return re.sub(regex, replacement, text, count=0, flags=re.MULTILINE)
  else:
    text = re.sub(r'\\', r'\\\\', text)
    return _MARKDOWN_ESCAPE_REGEX.sub(r'\\\1', text)


def escape_mentions(text: str) -> str:
  """A helper function that escapes everyone, here, role, and user mentions.

  .. note::

      This does not include channel mentions.

  .. note::

      For more granular control over what mentions should be escaped
      within messages, refer to the :class:`~discord.AllowedMentions`
      class.

  Parameters
  -----------
  text: :class:`str`
      The text to escape mentions from.

  Returns
  --------
  :class:`str`
      The text with the mentions removed.
  """
  return re.sub(r'@(everyone|here|[!&]?[0-9]{17,20})', '@\u200b\\1', text)



import subprocess
import sys

from .tcr_console import console as c


class Dependency:
  import_name: str
  pip_name: str

  def __init__(self, *, import_as: str, pip_install_as: str = None) -> None:
    self.import_name = import_as
    self.pip_name = pip_install_as or import_as

  def install(self) -> bool:
    """Returns bool if the installation succeeded or not."""
    try:
      subprocess.run([sys.executable, '-m', 'pip', 'install', self.pip_name], check=True)
    except subprocess.CalledProcessError:
      return False
    except Exception:
      return False
    else:
      return True


class DependencyEnsurer:
  deps: tuple[Dependency]
  auto_install: bool

  def __init__(self, *deps: Dependency, auto_install: bool = False) -> None:
    """Allows libraries to have optional dependencies (I KNOW OF `lib[extra]` syntax..! - you should probably use that thing instead).

    Requires at least 1 dependency to work. (will do nothing if `deps=()`)

    auto_install: If true, will install any missing dependencies automatically, if False, a RuntimeError will be raised and a message will be displayed. Warning: This may heavily slow down the program and may crash if no internet connection is present to fetch the missing dependencies.
    """
    self.deps = deps
    self.auto_install = auto_install

  def __call__(self, auto_install_override: bool | None = None) -> bool:
    """Verify if this package's dependencies are installed."""

    auto_install = auto_install_override if auto_install_override is not None else self.auto_install

    missing_deps: list[Dependency] = []

    for dep in self.deps:
      try:
        __import__(dep.import_name)
      except ImportError:
        missing_deps.append(dep)

    def errored():
      c.error('The following dependencies are missing:')
      for dep in missing_deps:
        c.error(f'{dep.import_name!r} (pip install {dep.pip_name})')
      raise RuntimeError('Missing dependencies (see errors above for tips!!).')

    if missing_deps:
      if auto_install:
        for dep in missing_deps.copy():
          if not dep.install():
            errored()
          else:
            missing_deps.remove(dep)
      else:
        errored()



from .. import util as __


@__.REQUIRE('__vars')
def var(_, name: str, value: str | None = None, *args, __vars: dict[str, str], **kwargs):
  """### Simple, single-scope variables.

  Two modes:
   - Set: `{var|key|value}`
   - Get: `{var|key}` OR `{key}`

  You first have to set a value to something - if it doesn't exist `''` (an empty string) is returned.

  Example:
  ```txt
  {var|var1|nya}{var1} {var|var1}
  ```

  Result:
  ```txt
  nya nya
  ```
  """
  if value is not None:
    __vars[name] = value
  else:
    return __vars.get(name, '')


@__.REQUIRE_POSITIONAL(2, None)
def test(_, left: str, right: str, iftrue: str = 'true', iffalse: str = 'false', **ctxs) -> str:
  """### =IF() from excel. Returns the third parameter if the first two are equal otherwise the fourth.

  If less than 2 parameters are passed, the placeholder is left uneavaluated (stays as a string of '{test|arg1})

  Example:
  ```txt
  {test| {username} | Colon | Your username is Colon! | Your username is not Colon!}
  ```

  Result:
  ```txt
  Your username is not Colon!
  (but if Colon ran this command, it would give the other response!)
  ```
  """
  if left == right:
    return iftrue
  else:
    return iffalse


def comment(*args, **ctxs) -> str:
  """### Return empty string no matter the arguments/contexts.

  Example:
  ```txt
  {#|this is a comment}This is te{#|another comment}xt
  ```

  Result:
  ```txt
  This is text
  ```
  """


__all__ = [x for x in globals() if not x.startswith('_')]



import functools
import math
import typing as t

if t.TYPE_CHECKING:
  import arc
import re as regex
from collections.abc import Callable
from enum import Enum, auto
from functools import wraps

import hikari

from ..src.tcr_compare import able
from ..src.tcr_console import console as c
from ..src.tcr_int import hex as tcrhex
from ..src.tcr_regex import RegexPreset
from ..src.tcr_run import run_sac
from .error import *


class _TT(Enum):
  TEXT = auto()
  PAREN_OPEN = auto()
  PAREN_CLOSE = auto()


class _Token:
  def __init__(self, type: _TT, value: str | None = None) -> None:
    self.type = type
    self.value = value

  def __repr__(self) -> str:
    return ('%s: %s' % (self.type.name, repr(self.value))) if self.value is not None else ('%s' % self.type.name)  # noqa: UP031


def _builder(x: str | bool) -> '_Token':
  if x is True:
    return _Token(_TT.PAREN_OPEN)
  elif x is False:
    return _Token(_TT.PAREN_CLOSE)
  elif isinstance(x, str):
    return _Token(_TT.TEXT, str(x))
  else:
    raise DynamicResponseSyntaxError(f'Unknown token type requested on build_placeholder_rich_return_value(): {x!r}')


def build_placeholder_rich_return_value(*args: str | bool) -> list[_Token]:
  """Build a value returnable from a placeholder callable, which can be used to insert a placeholder instead of a string.

  This does not validate input.
  - str: TEXT token
  - True: PAREN_OPEN token
  - False: PAREN_CLOSE token
  """
  return [_builder(x) for x in args]


def REQUIRE(*ctxs: str):
  """Require at least 1 of the given contexts to be present, otherwise raise DynamicResponsePlaceholderMissingContextError."""

  def decorator(func):
    @wraps(func)
    async def wrapper(*args, **kwargs):
      for ctx in ctxs:
        if ctx in kwargs and kwargs[ctx] is not None:
          for ctx in ctxs:
            if ctx not in kwargs:
              kwargs[ctx] = None
          return await run_sac(func, *args, **kwargs)
      raise DynamicResponsePlaceholderMissingContextError(f'One of {ctxs!r} must be present')

    return wrapper

  return decorator


def REQUIRE_AUTHOR(func):
  @REQUIRE('event', 'ctx')
  @wraps(func)
  async def wrapper(*args, event, ctx, **kwargs):
    return await run_sac(func, *args, author=(ctx.author if ctx else event.author), event=event, ctx=ctx, **kwargs)

  return wrapper


def SWITCH(switch_name: str, *, remove: bool = True):
  """Append a bool context depending on whether or not the passed phrase is in args. Optionally remove it."""

  def decorator(func):
    @wraps(func)
    async def wrapper(*args, **kwargs):
      if switch_name in args:
        if remove:
          args = tuple(x for x in args if x != switch_name)
        kwargs[switch_name] = True
      else:
        kwargs[switch_name] = False
      return await run_sac(func, *args, **kwargs)

    return wrapper

  return decorator


def DEBUG_CATCH_ERRORS(func):
  @wraps(func)
  async def wrapper(*args, **kwargs):
    try:
      return await run_sac(func, *args, **kwargs)
    except Exception as e:
      print(e)
      input()
      raise

  return wrapper


def REQUIRE_POSITIONAL(required_positionals: int, default: str | None = None):
  def decorator(func):
    @wraps(func)
    async def wrapper(*args, **kwargs):
      if len(args) < (required_positionals + 1):
        return default if default is not None else rebuild_yourself(*args, **kwargs)
      return await run_sac(func, *args, **kwargs)

    return wrapper

  return decorator


def STRINGIFY(func: Callable):
  @wraps(func)
  async def wrapper(*args, **kwargs):
    return str(await run_sac(func, *args, **kwargs))

  return wrapper


def FLATTEN_IF_POSSIBLE(func: Callable):
  @wraps(func)
  async def wrapper(*args, **kwargs):
    return flatten_if_possible(await run_sac(func, *args, **kwargs))

  return wrapper


def REQUIRE_GUILD(default: Callable[..., str] | str = ''):
  def decorator(func):
    @REQUIRE('event', 'ctx')
    @wraps(func)
    async def wrapper(*args, ctx, event, **kwargs):
      if ctx:
        guild = ctx.get_guild()
      elif hasattr(event, 'get_guild'):
        guild = event.get_guild()
      else:
        guild = None

      if guild is None:
        return default(func, args, ctx=ctx, event=event, **kwargs) if callable(default) else default

      return await run_sac(func, *args, guild=guild, ctx=ctx, event=event, **kwargs)

    return wrapper

  return decorator


def REQUIRE_MEMBER_AUTHOR(default: Callable[..., str] | str = ''):
  def decorator(func):
    @REQUIRE_AUTHOR
    @REQUIRE_GUILD(default)
    @wraps(func)
    async def wrapper(*args, guild, author, **kwargs):
      return await run_sac(func, *args, member=guild.get_member(author), guild=guild, author=author, **kwargs)

    return wrapper

  return decorator


def rebuild_yourself(*args: str, __parens: tuple[str, str], __splitter: str, **_):
  """# Rebuild Yourself (RYS).

  Return the placeholder as text, as if it wasnt evaluated at all.
  For example {dummy} -> returns the str '{dummy}'. Therefore it looks like the placeholder wasnt evaluated.

  This requires at least __parens and __splitter built-in contexts but can take entire **contexts and void others.

  This also supports argumented placeholders for example: {div|1|0} # Division by 0 is invalid so your implementation may be to return the str '{div|1|0}' - although that's not a very good example (it'd be better to return some numeric value for consistency) it's one that ilustrates the point.
  """
  return f'{__parens[0]}{(__splitter.join(args))}{__parens[1]}'


def flatten_if_possible(n: float | int) -> float | int:
  if n.is_integer():
    return int(n)
  return n


def number(
  s: str,
  *,
  flatten_to_int_when_possible: bool = False,
  default: int | float = 0,
) -> float | int:
  if r := able(float, s):
    a = float(r.result)
    if flatten_to_int_when_possible:
      return flatten_if_possible(a)
    else:
      return a
  return default


def jsbool(b: bool) -> t.Literal['true', 'false']:
  return 'true' if b else 'false'



from collections.abc import Callable
from typing import Any


def insist(
  func: Callable[[], Any],
  validator: Callable[[Any], bool] = bool,
  func2: Callable[[], Any] | None = None,
):
  """### Keep invoking `func` until `validator(func())` returns Truey result, return it.

  # Use `functools.partial` for this!!!

  ```py
  from functools import partial
  number = int(insist(
    partial(input, "Input a number: "),
    partial(tcr.able, int)
  ))
  ```

  `func2` is invoked on 2nd and every later iteration, unless not provided, in that case `func` is invoked at all times.

  Keep in mind that insist won't convert the, str that was returned by `input()` in the above example, only verify its compliance with validator.
  """
  result = func()
  if validator(result):
    return result
  while True:
    result = (func2 if func2 else func)()
    if not validator(result):
      continue

    return result



from collections.abc import Callable, Coroutine
from typing import Any, NotRequired, TypedDict, Unpack

import hikari
import miru

from ..src.tcr_null import Null


def embed(
  title: str,
  description: str,
  *,
  url=None,
  color=None,
  timestamp=None,
  thumbnail=None,
  footer=None,
  footer_icon=None,
  author: dict | None = None,
  image=None,
  fields: list | None = None,
) -> hikari.Embed:
  if (title is not Null and not title.strip()) or (description is not Null and not description.strip()):
    msg = f'Both title and description must be non-whitespace-only strings unless explicitly specified the title to be Null, got Title: {title!r}, Description: {description!r}'
    raise ValueError(msg)

  if fields is None:
    fields = []

  if author is None:
    author = {}

  out = hikari.Embed(
    title=title if title is not Null else None,
    description=description if description is not Null else None,
    color=color,
    timestamp=timestamp,
    url=url,
  )

  if thumbnail:
    out = out.set_thumbnail(thumbnail)

  if footer:
    out = out.set_footer(text=footer, icon=footer_icon)
  if author:
    out = out.set_author(**author)
  if image:
    out = out.set_image(image)

  for field in fields:
    if len(field) == 2:
      field = (*field, False)
    out = out.add_field(field[0], field[1], inline=field[2])
  return out


class ModalKwargs(TypedDict):
  title: str
  custom_id: NotRequired[str | None]
  timeout: NotRequired[str | None]


async def modal(
  responder: Callable[[miru.Modal], Coroutine[Any, Any, None]],
  callback: Callable[[miru.Modal, miru.ModalContext, list[str]], Coroutine[Any, Any, None]],
  *fields: miru.TextInput,
  **modal_kwargs: Unpack[ModalKwargs],
) -> dict[str, str]:
  """Create modal with passed fields, respond with it with passed responder (ctx.respond_with_modal), then return dict[field: str, user_input: str]."""

  class Modal(miru.Modal):
    async def callback(self, ctx: miru.ModalContext) -> None:
      await callback(self, ctx, list(self.values.values()))

  modal = Modal(**modal_kwargs)

  for field in fields:
    modal.add_item(field)

  await responder(modal)



from collections.abc import Callable, Iterable
from typing import Any

from .tcr_dict import merge_dicts


def dir2(x: object, /, dir: Callable[[object], Iterable[str]] = dir) -> list[str]:
  """Same as dir, but exclude the entries starting with '__' (double underscore)."""
  return [y for y in dir(x) if not y.startswith('__')]

def vars2(x: object, /, vars: Callable[[object], dict[str, Any]] = vars) -> dict[str, Any]:
  """Same as vars, but exclude the keys starting with '__' (double underscore)."""
  return {k: v for k, v in vars(x).items() if not k.startswith('__')}

def dir3(x: object, /, dir: Callable[[object], Iterable[str]] = dir) -> list[str]:
  """Same as dir, but exclude the entries starting with '_' (single underscore)."""
  return [y for y in dir(x) if not y.startswith('_')]

def vars3(x: object, /, vars: Callable[[object], dict[str, Any]] = vars) -> dict[str, Any]:
  """Same as vars, but exclude the keys starting with '_' (single underscore)."""
  return {k: v for k, v in vars(x).items() if not k.startswith('_')}

def dir_recursive(x: object, /) -> list[str]:
  """Checks all attributes recursively and returns a kind of recursive dir(). Also picks up on defaults since it goes all the way down to type/object in mro."""
  return list(vars_recursive(x))

def vars_recursive(x: object, /) -> dict[str, Any]:
  """Checks all attributes recursively and returns a kind of recursive vars(). Also picks up on defaults since it goes all the way down to type/object in mro."""
  return merge_dicts(x.__dict__, *[o.__dict__ for o in x.__class__.__mro__], {})



from functools import wraps

from .tcr_error import error


def apostrophe_s(word: str):
  """### Adds 's to a word unless it ends with s, then it adds only the apostrphe '.

  ```py
  >>> apostrophe_s('Mark')
  "Mark's"
  >>> apostrophe_s('James')
  "James'"
  """
  if word.endswith('s'):
    return word + "'"
  return word + "'s"


def plural_s(n: int, *, s: str = 's'):
  """Return `'s'` if `n != 1` else `''`."""
  return s if n != 1 else ''


def _caps_deco(func):
  @wraps(func)
  def wrapper(word: str, *args, **kwargs):
    result = func(word, *args, **kwargs)
    if word.isupper():
      return result.upper()
    if word[0].isupper() and word[1:].islower():
      return result[0].upper() + result[1:].lower()
    return result

  return wrapper


irregular_plurals = {
  'child': 'children',
  'man': 'men',
  'woman': 'women',
  'tooth': 'teeth',
  'foot': 'feet',
  'goose': 'geese',
  'mouse': 'mice',
  'ox': 'oxen',
  'person': 'people',
  'die': 'dice',
  'criterion': 'criteria',
  'phenomenon': 'phenomena',
  'cactus': 'cacti',
  'focus': 'foci',
  'analysis': 'analyses',
  'thesis': 'theses',
  'crisis': 'crises',
  'basis': 'bases',
  'stimulus': 'stimuli',
  'appendix': 'appendices',
  'bacterium': 'bacteria',
  'curriculum': 'curricula',
  'index': 'indices',
  'medium': 'media',
  'millennium': 'millennia',
  'radius': 'radii',
  'memorandum': 'memoranda',
  'vertex': 'vertices',
  'belief': 'beliefs',
  'chief': 'chiefs',
  'roof': 'roofs',
}


def make_plural(word: str, n: int = 2) -> str:
  """### Tries its best to convert the word to plural if n != 1 else return the original word."""
  if n == 1:
    return word
  if word.lower() in irregular_plurals:
    return irregular_plurals[word.lower()]
  if word.lower().endswith(('s', 'x', 'z', 'ch', 'sh', 'o')):
    return word + 'es'
  if word.lower().endswith('y') and word[-2] not in 'aeiou':
    return word[:-1] + 'ies'
  if word.lower().endswith('f'):
    return word[:-1] + 'ves'
  if word.lower().endswith('fe'):
    return word[:-2] + 'ves'
  return word + 's'


globals()['make_plural'] = _caps_deco(make_plural)


def nth(n: int):
  """### Return a string containing original number + numeric suffix (st, nd, rd, th).

  Takes into account edge cases like 11, 12.
  Supports negative numbers & zero.

  Examples:
  ```py
  >>> nth(1)
  '1st'
  >>> nth(4)
  '4th'
  >>> nth(12)
  '12th'
  ```
  """
  if not isinstance(n, int):
    raise error.NotIntegerError(n)

  if n < 0:
    return '-' + nth(-n)

  if 10 < n % 100 < 14:
    suffix = 'th'
  else:
    match n % 10:
      case 1:
        suffix = 'st'
      case 2:
        suffix = 'nd'
      case 3:
        suffix = 'rd'
      case _:
        suffix = 'th'

  return str(n) + suffix



from typing import Literal, NotRequired, TypedDict

# TODO: Maybe add a parse_profile() function for parsing a "No profiles" into a ListedProfileError.
# TODO: Maybe add a parse_profile() function for parsing a "No profiles" into a ListedProfileError.
# TODO: Maybe add a parse_profile() function for parsing a "No profiles" into a ListedProfileError.
# TODO: Maybe add a parse_profile() function for parsing a "No profiles" into a ListedProfileError.
# TODO: Maybe add a parse_profile() function for parsing a "No profiles" into a ListedProfileError.


class ListedProfileError(TypedDict):
  """Represents an error returned from /api/profiles.

  TODO: review this..., surely there are more fields? wait no it just returns a "No profiles" (fix this)
  """

  error: Literal[True]


class ListedProfileIconTD(TypedDict):
  """A dictionary of the user's icon options: `emoji` and `parsed`."""

  emoji: str
  """Represents the emoji form of that user's icon, for use on Discord."""
  parsed: str
  """Represents HTML version of that emoji, for use outside of Discord."""


class ListedProfile(TypedDict):
  """Not to be confused with Profile.

  Represents a dict of a single user profile received from /api/profiles, plural (not to confuse with /api/profile, singular)
  """

  id: str
  """Discord ID of that user along with the internal profile ID, example: `'507642999992352779_ox'`."""
  name: str
  """Name of that user's zoo."""
  color: int
  """Color of that user's embed."""
  private: bool
  """Whether or not that zoo is private."""
  viewable: bool
  """Whether or not that zoo is viewable."""
  current: bool
  """Whether or not that user's zoo is the currently selected one."""
  score: int
  """Zoo score of that user's profile."""
  icon: ListedProfileIconTD


class ProfileError(TypedDict):
  """Represents a an error returned from /api/profile.

  This does not cover network-related issues. It's purely for errors caught by Colon's servers.
  """

  name: str
  """For example: "Invalid Profile!".

  TODO: make it a Literal["...", "...", "..."].
  """
  msg: str
  """The error message."""
  invalid: bool
  """TODO: idk what that does."""
  error: str
  """Type of the error. for example: "invalidProfile"."""
  login: NotRequired[bool]
  """TODO: find out what that does exactly. (guess: true if if you logged in then you'd be able to access this profile)"""


class ProfileUserTD(TypedDict):
  """A dictionary of the user's info: `avatar`."""

  avatar: str
  """Represents the URL to discord's cdn pointing to that user's avatar"""


class ProfileUniqueAnimalsTD(TypedDict):
  """A dictionary of the user's unique animals: `common`, `rare`, `total`."""

  common: int
  """Represents unique common animals."""
  rare: int
  """Represents unique rare animals."""
  total: int
  """Total of animal pairs (common-rare) of which at least one of them (either common or rare) is present."""


class ProfileTotalAnimalsTD(TypedDict):
  """A dictionary of the user's total animals: `common`, `rare`."""

  common: int
  """Represents total common animals."""
  rare: int
  """Represents total rare animals."""


class ProfileAnimalTD(TypedDict):
  """Represents a single animal group from the user's profile."""

  name: str
  """The name of that animal."""
  amount: int
  """The amount of that animal."""
  emoji: str
  """The emoji of that animal."""
  emojiName: str
  """The name of emoji of that animal."""
  family: str
  """The family of that animal."""
  rare: bool
  """True if rare, False if common."""
  pinned: bool
  """Whether or not user used /pin on that animal on that profile."""


class ProfileItemTD(TypedDict):
  name: str
  """The name of that item."""
  amount: int
  """The amount of that item."""
  emoji: str
  """The emoji of that item."""
  emojiName: str
  """The name of emoji of that item."""
  highlight: bool
  """Whether or not this item should be highlighted (is upgraded?)."""
  description: str | None
  """The description of that item.

  Seems to be gone after some recent-ish update (it's always None now), may return if you provide appropriate headers to prove that profile is yours.
  """
  timesUsed: int
  """The amount of times that item has been used on that profile."""


class ProfileRelicTD(TypedDict):
  name: str
  """The name of that relic."""
  emoji: str
  """The emoji of that relic."""
  description: str | None
  """The description of that relic.

  Seems to be gone after some recent-ish update (it's always None now), may return if you provide appropriate headers to prove that profile is yours.
  """


class ProfileCosmeticTD(TypedDict):
  name: str
  """The name of that cosmetic."""
  emoji: str
  """The emoji of that cosmetic."""
  trophy: NotRequired[Literal[1, 2, 3]]
  """Whether or not that cosmetic is a trophy.

  - This cosmetic is NOT a trophy if THIS KEY IS NOT PRESENT
  - This cosmetic is a common trophy if the value is 1
  - This cosmetic is a rare trophy if the value is 2
  - This cosmetic is a leader trophy if the value is 3
  """


class ProfileQuestsTD(TypedDict):
  """Represents a list of quests from the user's profile. Not to be confused with `ProfileQuestTD`."""

  name: str
  """The name of that quest. (ex. 'Risky Quest')"""
  type: str
  """The internal name of that quest (ex. 'risky')"""
  emoji: str
  """The emoji for that quest. (This should be a unicode emoji 100% of the time but colon may or may not add other quests that make use of custom emojis so don't rely on this being a printable string maybe)"""
  days: int
  """How manydays does this quest take to complete."""
  completed: int
  """How many quests of that type were completed on that profile."""


class ProfileQuestTD(TypedDict):
  """Represents the currently pending quest on that user's profile. Not to be confused with `ProfileQuestsTD`."""

  type: str
  """The internal name of that quest (ex. 'risky')."""
  animal: str
  """Name of the rare animal that went on this quest. (ex. 'Wolf')."""
  family: str
  """The family of the animal that went on this quest (ex. 'fox')."""


class ProfileCurseNamesTD(TypedDict):
  type: str
  """Type of the curse."""
  cure: str
  """Cure of the curse."""


class ProfileCurseEffectsTypeTD(TypedDict):
  name: str
  """Name of the curse. Ex: `'Blindness'`."""
  description: str
  """Description of the curse. Ex: `'Rescued animals are no longer revealed.'."""
  weak: None
  """TODO: idk what it's exactly, seems to be able to be None."""


class ProfileCurseEffectsCureTD(TypedDict):
  name: str
  """Name of the curse's cure. Ex: `'Shackled'`."""
  description: str
  """Description of the curse. Ex: `'Cannot be removed. Rescue luck is drastically increased while under this curse.'."""
  weak: None
  """TODO: idk what it's exactly, seems to be able to be None."""


class ProfileCurseEffectsTD(TypedDict):
  type: ProfileCurseEffectsTypeTD
  """Type of the curse."""
  cure: ProfileCurseEffectsCureTD
  """Cure of the curse."""


class ProfileCurseTD(TypedDict):
  name: str
  """The full name of the curse. Ex: `'Shackled Curse of Blindness'`."""
  names: ProfileCurseNamesTD
  """Parsed curse details (`type` and `cure`)."""
  weak: bool
  """Whether or not the curse has been weakened."""
  effects: ProfileCurseEffectsTD


class ProfileTerminalFishyTD(TypedDict):
  commonFish: int
  uncommonFish: int
  rareFish: int
  trash: int
  pebbles: int


class ProfileTerminalGardenTD(TypedDict):
  unlocked: bool
  """Whether or not the garden is unlocked."""


class ProfileTerminalCardsTD(TypedDict):
  total: int
  """Total amount of cards."""
  common: int
  """Amount of common cards."""
  rare: int
  """Amount of rare cards."""
  ultraRare: int
  """Amount of legendary cards."""


class ProfileTerminalFusionFusionsTD(TypedDict):
  commonCommon: int
  commonRare: int
  rareRare: int
  total: int
  score: int


class ProfileTerminalFusionNFBsTD(TypedDict):
  common: int
  rare: int
  total: int
  score: int


class ProfileTerminalFusionTD(TypedDict):
  tokensPerRescue: int
  tokensFromFusions: int
  nfbMultiplier: float | int
  """I think that's float? but it also returns int sometimes??"""
  fusions: ProfileTerminalFusionFusionsTD
  nfbs: ProfileTerminalFusionNFBsTD


class ProfileTerminalTD(TypedDict):
  unlocked: bool
  """Whether or not the terminal is unlocked."""
  admin: bool
  """Whether or not the user is a terminal administrator on this profile."""
  commandsFound: int
  """Amount of discovered discoverable terminal commands on this profile."""
  mechanicPoints: NotRequired[int]
  """Murphy points, used in /mechanic. You should know this if you're that far into zoo to be reading a zoo API wrapper docs... 👀"""
  fishy: NotRequired[ProfileTerminalFishyTD]
  """Stats from the terminal's `$ fishy` minigame."""
  garden: NotRequired[ProfileTerminalGardenTD]
  """Stats from the terminal's `$ garden`."""
  cards: NotRequired[ProfileTerminalCardsTD]
  """Stats from the terminal's cards"""
  fusion: NotRequired[ProfileTerminalFusionTD]


class ProfileGoalTD(TypedDict):
  name: str
  """The name of the goal. Ex: `'Zookeeper'`."""
  emoji: str
  """The emoji of the goal. This can be a custom emoji so don't rely on it being printable."""
  tier: str
  """The tier formatted as a roman numeral string Ex: `'VIII'`."""
  tierNumber: int
  """Same as tier but a python integer, not a roman numeral string.."""
  target: int
  """Target amount of the goal objective."""
  desc: str
  """Description of the goal, Ex: `'Reach 1,000✧ zoo score'`"""
  count: int
  """The actual amount of the goal objective."""
  complete: bool
  """Whether or not the goal is complete."""


class ProfileSettingsTD(TypedDict):
  altTimestamp: bool
  """Whether or not to display timestamps in an alternate format."""
  fastConfirmations: bool
  """Whether or not to skip some confirmations."""
  showAnimalTotals: bool
  """Whether or not to show animal totals on rescue."""
  disableNotifications: bool
  """Whether or not to disable automatic use of notifications."""
  disableAutoRescues: bool
  """Whether or not to disable automatic use of auto rescues."""
  disableQuestNotifications: bool
  """Whether or not to disable quest notifications."""
  disableCustomColor: bool
  """Whether or not to disable custom colors."""
  hideCosmetics: bool
  """Whether or not to hide cosmetics to other users."""


class ProfileSecretInfoShopTD(TypedDict):
  credits: int
  """Amount of shop credits for that profile."""
  nextCredit: int
  """Unix timestamp of when the next credit will be given. (not sure what happens if you're maxed out)"""
  maxCredits: int
  """Max amount of shop credits for that profile. (can be increased using Sack relic or something)"""
  lastPurchase: int  # | None?
  """Unix timestamp of when the last purchase was made. (I'm not sure what is it if you've never used the shop)"""


class ProfileSecretInfoCooldownsTD(TypedDict):
  """Contains unix timestamps for when the cooldowns will be over."""

  rescue: int
  relic: int
  leader: int
  profile: int  # I'm not sure why is it bound to a profile if it's also bound to a user.
  cardPull: int
  pet: int
  fishy: int
  sisyphus: int


class ProfileSecretInfoTerminalTD(TypedDict):
  directory: str
  """The current working directory in /terminal of that profile."""
  commands: list[str]
  """Full list of discovered discoverable commands in /terminal of that profile."""
  nextFusion: int
  """Amount of rescues needed for the next fusion"""


class ProfileSecretInfoGardenTD(TypedDict):
  nextPlant: int
  """The unix timestamp of when the next use of `$ garden plant` will be available."""
  watered: bool
  """Whether or not some plant within garden was watered since the last rescue."""
  longestPlant: int
  """Age of the longest living plant in miliseconds."""
  sprinkler: NotRequired[Literal[1, 2, 3]]
  """Currently selected slot of the sprinkler.

  This key is not present if sprinkler is not equipped.
  """


class ProfileSecretInfoTD(TypedDict):
  sort: int
  """Selected sorting. TODO: find out which values are which sortings."""
  color: int
  """The color of uh... what is this for? the color is already public? huh? idk (TODO: find out)."""
  promise: dict
  """Promises made to the user on that profile (for example: next rescue aquatic type)."""
  questEnd: int | None
  """Unix timestamp of when the quest ends. May be None if no quest is active."""
  questBoosts: dict
  """(guess) Items like telescope used on that quest."""
  curseEnd: int | None
  """Unix timestamp of when the curse ends. May be None if no curse is active."""
  mechanicEnd: int | None
  """Unix timestamp of when the mechanic upgrade ends. May be None if no mechanic upgrade is active."""
  shop: ProfileSecretInfoShopTD
  """/shop-related information about that profile."""
  cooldowns: ProfileSecretInfoCooldownsTD
  """Contains unix timestamps for when the cooldowns will be over."""
  terminal: ProfileSecretInfoTerminalTD
  """/terminal-related information about that profile."""
  garden: ProfileSecretInfoGardenTD


class Profile(TypedDict):
  """Not to be confused with ListedProfile.

  Represents information about profile received from /api/profile, singular (not to confuse with /api/profiles, plural)
  """

  id: int
  """The ID of that profile. Example: `1234123412341234_fox`."""
  userID: str
  """The Discord ID of that profile's user. Example: `1234123412341234`."""
  profileID: str
  """The internal ID of that profile. Example: `fox`."""
  selectedProfile: str
  """The profileID of the currently selected profile. Example: `fox`."""
  profiles: list[str]
  """List of all profiles of that user. Example: `['fox', 'dog', 'cat']`."""
  user: ProfileUserTD
  """Currently contains just key 'avatar' with the value that is a URL to discord's cdn pointing to that user's avatar."""
  name: str
  """The user-selected name of that zoo. (Not the user's discord name, the name of the zoo itself, not the user)."""
  nickname: str
  """The user-selected nickname of themselves, chosen with terminal's `$ nick` command."""
  color: str
  """The hex representation of that profile's embed color."""
  owner: bool
  """TODO: what does this do? (i guess it's whether or not user logged in on gdcolon.com and that's their zoo?)."""
  private: bool
  """Whether or not that zoo is private."""
  profileTheme: str
  """The selected profile theme internal ID, example: `'colon'`."""
  score: int
  """The zoo score of that profile."""
  completion: float
  """The completion percentage of that profile."""
  uniqueAnimals: ProfileUniqueAnimalsTD
  """That profile's unique animals: `common`, `rare`, `total`."""
  totalAnimals: ProfileTotalAnimalsTD
  """That profile's total animals: `common`, `rare`."""
  totalItems: int
  """That profile's total number of items."""
  totalCosmetics: int
  """That profile's total number of cosmetics."""
  totalTrophies: int
  """That profile's total number of trophies."""
  totalLeaderXP: int
  """That profile's total number of leader xp."""
  unspentLeaderXP: int
  """That profile's unspent leader xp."""
  equippedRelics: list[str]
  """List of equipped relics, example: `['Decearing Egg']`."""
  equippedCosmetics: list[str]
  """The full list of equipped cosmetics (for the emoji cosmetic see `equippedCosmetic`)"""
  equippedCosmetic: str
  """The cosmetic equipped as emoji. (for full list see `equippedCosmetics`)"""
  equippedLeader: str
  """The currently equipped leader's name"""
  cosmeticIcon: str
  """The cosmetic emoji (or discord emoji syntax if custom emoji is used)"""
  notifications: int
  """The amount of this profile's notifs"""
  autoRescues: int
  """The amount of this profile's auto rescues"""
  animals: list[ProfileAnimalTD]
  """List of that profile's animals."""
  items: list[ProfileItemTD]
  """List of that profile's items."""
  relics: list[ProfileRelicTD]
  """List of that profile's relics."""
  cosmetics: list[ProfileCosmeticTD]
  """List of that profile's cosmetics."""
  quests: list[ProfileQuestsTD]
  """List of that profile's quests."""
  quest: ProfileQuestTD
  """The currently pending quest."""
  curse: None | ProfileCurseTD
  """The currently pending curse (may be None if no curse is present)."""
  terminal: ProfileTerminalTD
  """Terminal-related information about the profile."""
  stats: list
  """TODO: this seems to always return an empty list?."""
  goals: list[ProfileGoalTD]
  """A full list of goals with details. For goal tier number see `goalTiers`."""
  goalTiers: int
  """The number of goal tiers completed. For a full list of goals with details see `goals`."""
  extraData: list[list[str, str, NotRequired[int]]]
  """Extra information about the profile in form of little badges.

  element idx=0 (required): str = The emoji (sometimes but not always unicode) of the badge. (Ex: '<:fluff:1078401638370385960>')
  element idx=1 (required): str = The name of the badge. (Ex: 'Made the game')
  element idx=2 (optional): int = Some quantity of that badge, not included if not applicable. (Ex: 92; Context: if the previous values were associated with for example "pebbles")
  """
  settings: ProfileSettingsTD
  """This profile's current /settings."""
  secretInfo: NotRequired[ProfileSecretInfoTD]
  """This profile's so called secret info.

  This key is included if you have the [API Key](https://zoobot.wiki/index.php/API_Key) relic equipped.
  """



import contextlib
from collections.abc import Generator, Hashable, Mapping
from typing import Any

from .tcr_compare import isdunder
from .tcr_null import Undefined


def _check_strict(master: Mapping, slave: Mapping, *, recursive: bool) -> bool:
  for k, v in master.items():
    if k not in slave:
      return [k]
    if recursive and isinstance(v, Mapping) and isinstance(slave[k], Mapping):  # noqa: SIM102
      if (a := _check_strict(v, slave[k], recursive=recursive)) is not True:
        return [k, *a]
  return True


def dict_zip(dict1: Mapping, *dicts: Mapping) -> Generator[tuple[Hashable, Any], None, None]:
  """Zips dictionaries simillar to `zip()`, returns a generator which yields a tuple of (key, (value1, value2, value3)) for each key. This function does not support analogy of `zip()` non-strict mode, strict=True is implied (Each key must be in every single passed in dict)."""

  dicks = [dict1, *dicts]

  kys = dicks[0].keys()
  if any(kys != dick.keys() for dick in dicks):
    msg = 'Mismatched keys'
    raise ValueError(msg)

  for k, v in dicks[0].items():
    yield (k, (v, *(dick[k] for dick in dicks[1:])))


def merge_dicts(master: Mapping = None, slave: Mapping = None, *dicts: Mapping, recursive=True, strict=False) -> dict:
  """Merge dictionaries, made to prioritize the `master` dictionary and if key is not found there, then it takes from the `slave` dictionary.

  Optionally if `recursive=True`, then if the same key is a dict in both master and slave, merge dicts operation is performed on both of them.\\
  Optionally if `strict=True` it raises `ValueError` if there exists a key that is in master dictionary but not in slave dictionary.\\
  If there are more than 2 arguments passed in the list is reduced from back to front until two elements are left by the process of merge_dicts(master=list[-2], slave=list[-1])
  """
  merged = {}

  if master is None:
    master = {}
  if slave is None:
    slave = {}

  dicts = [master, slave, *dicts]

  while len(dicts) > 2:
    slave_ = dicts.pop()
    master_ = dicts[-1]
    dicts[-1] = merge_dicts(master_, slave_, recursive=recursive, strict=strict)

  master, slave = dicts

  if strict:
    a = _check_strict(master, slave, recursive=recursive)
    if a is not True:
      msg = f'Strict check failed: there exists a key ({".".join(a)}) that is in master dictionary but not in slave dictionary'
      raise ValueError(msg)

  for key in master:
    if key in slave:
      if isinstance(master[key], dict) and isinstance(slave[key], dict) and recursive:
        merged[key] = merge_dicts(master[key], slave[key], recursive=True)
      else:
        merged[key] = master[key]
    else:
      merged[key] = master[key]

  for key in slave:
    if key not in merged:
      merged[key] = slave[key]

  return merged


def clean_dunder_dict(
  __dict__: dict[str, Any],
  strategy: int = 1,
) -> dict:
  """Filter the __dict__ passed in (or any other str-keyed dict for that matter).

  Strategy types:
  - 0: Remove only __dunder__ keys
  - 1: Do whatever lower-number strategies do but also remove __double_underscore_prefixed keys (default)
  - 2: Do whatever lower-number strategies do but also remove _single_underscore_prefixed keys
  """

  if strategy == 0:
    return {x: y for x, y in __dict__.items() if not (x.startswith('__') and x.endswith('__'))}

  if strategy == 1:
    return {x: y for x, y in __dict__.items() if not x.startswith('__')}

  if strategy == 2:
    return {x: y for x, y in __dict__.items() if not x.startswith('_')}

  if strategy == -2:
    raise ValueError('Invalid strategy, choose 0, 1 or 2 (not literally `0-2` you fucking idiot)')

  raise ValueError('Invalid strategy, choose 0-2')


class DotDict(dict):
  """### dot.notation access to dictionary attributes.

  This will ignore any dunder attributes, and will not try to look them up in the underlying dict.

  ```py
  >>> d = {"a": 1, "b": 2}
  >>> dd = DotDict(d) # or DotDict({"a": ...})
  >>> dd.a
  1
  >>> dd.b
  2
  ```

  This will convert any Mapping that is not already a DotDict into a DotDict on access.

  Raises:
      - KeyError (tried to convert the dot-access to item-access but failed to find item)
      - AttributeError (the dot-accessed attribute was dunder, did not try to convert to item-access and said attribute was not found)
      - Other standard dict errors if any.
  """

  def __getattr__(self, attr):
    if isdunder(attr):
      return object.__getattribute__(self, attr)

    a = super().__getitem__(attr)

    if not isinstance(a, DotDict) and isinstance(a, Mapping):
      return DotDict(a)

    return a

  def __setattr__(self, attr, value):
    if isdunder(attr):
      return setattr(self, attr, value)

    super().__setitem__(attr, value)

  def __delattr__(self, attr):
    if isdunder(attr):
      return delattr(self, attr)

    super().__delitem__(attr)


class JSDict(dict):
  """Returns tcr.Undefined when a value is not found rather than erroring.

  In order to effectively replace the undefined value (tcr.Undefined) bitwise or it with any other value, (`tcr.Undefined | T -> T`).
  """

  def __getitem__(self, key):
    try:
      return super().__getitem__(key)
    except KeyError:
      return Undefined

  def __delitem__(self, __key: Any) -> None:
    with contextlib.suppress(KeyError):
      return super().__delitem__(__key)


class JSDotDict(DotDict, JSDict):
  """Combines tcr.DotDict and tcr.JSDict.

  (Allows for dot-access and returns tcr.Undefined when a value is not found rather than erroring.)

  For full description see the previously mentioned class doc strings.
  """



import inspect
import os
import pathlib as p
import re
from collections.abc import Callable

from .tcr_console import console


class path:
  """Provides path-related utilities."""

  @staticmethod
  def center(path: str | p.Path) -> p.Path:
    """### Change directory to the location of the passed in __file__ (or any string path or pathlib Path). This will NOT detect if the passed in location is a directory, it will go to parent of it instead. Return previous path as a pathlib.Path object.

    This function is shit use tcr.path.thisdir() instead.
    """
    a = os.getcwd()
    os.chdir(p.Path(path).resolve().parent)
    return p.Path(a)

  @staticmethod
  def nextname(name='New folder') -> str:
    """### If passed in name does not end with "`({int})" where `{int}` is any integer, return it with "` (1)`" appended else return it with the previously mentioned `{int}` incremented and leave it otherwise unchanged."""
    match = re.match(r'^(.*) \((\d+)\)$', name)

    if match:
      base_name = match.group(1)
      current_index = int(match.group(2))
      new_name = f'{base_name} ({current_index + 1})'
    else:
      new_name = f'{name} (1)'

    return new_name

  @staticmethod
  def nextname_file_ext_fix(name='New file.txt') -> str:
    match = re.match(r'^(.*?)(?:\s*\((\d+)\))?(?:\.(\w+))?$', name)
    if not match:
      raise ValueError('Invalid filename format')

    base_name, number, extension = match.groups()

    new_number = int(number) + 1 if number else 1

    return f'{base_name} ({new_number}).{extension}' if extension else f'{base_name} ({new_number})'

  @staticmethod
  def newdir(
    name: str | None = None,
    path: p.Path | str | None = None,
    nextname: Callable[[str | None], str] = nextname,
  ) -> str:
    """### Return unused directory or file with the name `name` in directory `path` or the current directory if not passed in."""
    if path is None:
      path = os.getcwd()
    path = str(path)
    listdir = os.listdir(path)
    while name in listdir:
      name = nextname(name)
    return name

  @staticmethod
  def thisdir(*, chdir=False) -> p.Path:
    """### Return the directory of the furthest back file in the callstack as pathlib.Path, optionally navigate to it with os.chdir()."""
    frames = inspect.stack()
    pth = frames[-1].filename  # Get the filename of the furthest back frame
    pth = p.Path(pth).parent
    if chdir:
      os.chdir(pth)
    return pth

  @staticmethod
  def nt_validname(nt_objname: str | p.Path) -> bool:
    """### Return True if the passed in `nt_objname` is a valid NT File/Folder/whatever, else return False.

    This does NOT check if the path exists.
    ### The passed in path is stripped(), the used one should be as well!
    """
    if not isinstance(nt_objname, str):
      nt_objname: str = str(nt_objname.absolute().name)

    return bool((all((x not in nt_objname) for x in ('\n', '\r'))) and (nt_objname.strip() not in ('', '.', '..')) and (re.match(r'^[^<>:"\/\\|?*\n]*$', nt_objname.strip())))


__all__ = ['path']



from builtins import hex as sex  # nocom (actually yescom because this text is a comment dummy!!)
from collections.abc import Iterable, Mapping
from numbers import Number


def hex(number: int, leading_zeroes=2, *, upper=True, prefix: str = '0x') -> str:
  """Slightly more advanced version of builtin `hex()`, offers ability to choose if uppercase and how many leading zeroes."""
  hex_output = sex(number)
  hex_value = hex_output[2:].zfill(leading_zeroes).upper() if upper else hex_output[2:].zfill(leading_zeroes)

  formatted_output = f'{prefix}{hex_value}'
  if not upper:
    formatted_output = formatted_output.lower()

  return formatted_output


def recursive_sum(it: Iterable) -> int | float:
  s = 0

  for item in it.values() if isinstance(it, Mapping) else it:
    s = s + recursive_sum(item) if isinstance(item, Iterable) else s + item

  return s


def float2int(n: float) -> float | int:
  if round(n) == n:
    return int(n)
  return float(n)


def clamp(min: Number, n: Number, max: Number) -> Number:
  return min if n < min else max if n > max else n



"""TCRUtils-specific errors."""
from types import TracebackType
from typing import Any


class error:
  class ConfigurationError(ValueError):
    """Used when a user-provided configuration causes an invalid state."""

  class NotIntegerError(TypeError):
    """Used when integer was expected but not supplied."""

    def __init__(self, not_integer: Any | None = None, /) -> None:
      s = 'Integer was expected.'
      if not_integer:
        s += f" Got '{not_integer!r} instead'"
      super().__init__(s)

  class DatabaseError(Exception):
    """Used when database raises an unexpected exception."""

    def __init__(self) -> None:
      super().__init__()

  class NoMatchingOverloadError(Exception):
    """Used in tcr.Overload."""


for attr_name, attr_value in error.__dict__.items():
  if not attr_name.startswith('__'):
    globals()[attr_name] = attr_value

__all__ = [x for x in globals() if not x.startswith('_') and x != error.__name__]



from .tcr_int import hex


class QuotelessString(str):
  """Used in tcr.fmt_iterable() in order to format it as a plain text, rather than as `"normal str"`."""


class HexInt(int):
  def __tcr_fmt__(self, fmt_iterable, **kwargs):
    return fmt_iterable(int(self), **{**kwargs, 'int_formatter': hex})


class UnreprableString(QuotelessString):
  def __repr__(self) -> str:
    return self



import time
from collections import OrderedDict
from typing import Any, Self


class Singleton:
  """Keep single instance at all times.

  ## Make sure to put the `tcr.Singleton` inheretance before all others!
  ### ❌ `class A(B, tcr.Singleton)`
  ### ✅ `class A(tcr.Singleton, B)`

  ```py
  class SingleTuple(tcr.Singleton, tuple):
      ...

    tup = SingleTuple(('tup1',))
    tup2 = SingleTuple(('tup2',)) # Those arguments are effectively ignored

    console(tup)  # -> ('tup1',)
    console(tup2) # -> ('tup1',)
  ```
  """

  __singleton_instance__ = None

  def __new__(cls, *args, **kwargs) -> Self:
    if cls.__singleton_instance__ is None:
      cls.__singleton_instance__ = super().__new__(cls, *args, **kwargs)
    return cls.__singleton_instance__


class NoInit:
  def __init__(self, *args, **kwargs) -> None:
    msg = 'You cannot instiantiate this class'

    if hasattribute(self, '__noinit_msg'):
      msg = str(getattr(self, '__noinit_msg'))

    raise RuntimeError(msg)


def partial_class(_f=None, /, *, add_as_kwargs_to_init: bool = True) -> type:
  """### Simillarly to functools.partial allows subclasses of decorated class to hold a meta-like state to be later used in e.g. root's init or anywhere else.

  ```py
  @tcr.partial_class
  class Button:
    def __init__(self, label: str):
      self.label = label

    def __repr__(self):
      return f'Button(label={self.label!r})'

  class YesButton(Button, label='Yes'): ...
  class NoButton(Button, label='No'): ...

  y = YesButton()
  c(repr(y)) # -> Button(label='Yes')
  n = NoButton()
  c(repr(n)) # -> Button(label='No')
  n2 = NoButton(label='Nuh uh') # Overridden
  c(repr(n2)) # -> Button(label='Nuh uh')
  ```

  Args:
    add_as_kwargs_to_init: bool = True. If True, the state  will be passed as keyword arguments into `__init__`.
  """

  def decorator(cls: type) -> type:
    def new_init_subclass(new_cls, **kwargs) -> None:
      new_cls._partial_state = kwargs

      if not add_as_kwargs_to_init:
        return

      original_init = new_cls.__init__

      def new_init(self, *args, **init_kwargs):
        merged_kwargs = {**self._partial_state, **init_kwargs}  # Merge both states such that caller can override it, like functools.partial
        original_init(self, *args, **merged_kwargs)

      new_cls.__init__ = new_init

    cls.__init_subclass__ = classmethod(new_init_subclass)
    return cls

  if _f is not None:
    return decorator(_f)
  return decorator


def hasattribute(obj: object, name: str) -> bool:
  try:
    object.__getattribute__(obj, name)
  except AttributeError:
    return False
  else:
    return True


class DefaultsGetSetItem(dict):
  def __getitem__(self, __key: Any) -> Any:
    try:
      return super().__getitem__(__key)
    except KeyError:
      if hasattribute(self, 'defaults') and __key in (defaults := object.__getattribute__(self, 'defaults')):
        val = defaults[__key]()
        self[__key] = val
        return val
      else:
        raise


class DefaultsGetItem(dict):
  def __getitem__(self, __key: Any) -> Any:
    try:
      return super().__getitem__(__key)
    except KeyError:
      if hasattribute(self, 'defaults') and __key in (defaults := object.__getattribute__(self, 'defaults')):
        return defaults[__key]()
      else:
        raise


class DefaultsGetSetAttr:
  def __getattr__(self, __name: str) -> Any:
    try:
      return getattr(super(), __name)
    except AttributeError as e:
      if hasattribute(self, 'defaults') and __name in (defaults := object.__getattribute__(self, 'defaults')):
        val = defaults[__name]()
        self.__setattr__(__name, val)
        return val
      else:
        raise AttributeError(f'{self.__class__.__name__!r} object has no attribute nor default value for {__name!r}') from e


class DefaultsGetAttr:
  def __getattr__(self, __name: str) -> Any:
    try:
      return getattr(super(), __name)
    except AttributeError as e:
      if hasattribute(self, 'defaults') and __name in (defaults := object.__getattribute__(self, 'defaults')):
        return defaults[__name]()
      else:
        raise AttributeError(f'{self.__class__.__name__!r} object has no attribute nor default value for {__name!r}') from e


class CachedInstancesMeta(type):
  """Metaclass for caching instances based on init parameters with customizable max time and max instances."""

  def __new__(
    mcls,  # noqa: N804
    name,
    bases,
    dct,
    max_instances: int = 50,
    max_time: int = 30 * 60,
    restore_method: str | None = None,
  ):
    cls = super().__new__(mcls, name, bases, dct)
    cls._cache = OrderedDict()
    cls._max_instances = max_instances
    cls._max_time = max_time
    cls._restore_method = restore_method
    return cls

  def __call__(cls, *args, **kwargs):
    key = (args, tuple(sorted(kwargs.items())))

    cls._remove_old_instances()

    if key in cls._cache:
      instance, timestamp = cls._cache.pop(key)
      cls._cache[key] = (instance, time.time())

      if instance._restore_method is not None:
        getattr(instance, instance._restore_method)()
      return instance

    instance = super().__call__(*args, **kwargs)

    cls._cache[key] = (instance, time.time())

    cls._enforce_max_cache_size()

    return instance

  def _remove_old_instances(cls):
    current_time = time.time()
    keys_to_remove = []
    for key, (_, timestamp) in cls._cache.items():
      if current_time - timestamp > cls._max_time:
        keys_to_remove.append(key)
    for key in keys_to_remove:
      del cls._cache[key]

  def _enforce_max_cache_size(cls):
    while len(cls._cache) > cls._max_instances:
      cls._cache.popitem(last=False)



import tempfile
from pathlib import Path


def temp_file(content: str = '', *, suffix: str = None) -> Path:
  temp_file = tempfile.NamedTemporaryFile(delete=False, mode='w', encoding='utf-8', suffix=suffix)

  try:
    temp_file.write(content)
    temp_file_path = Path(temp_file.name)
  finally:
    temp_file.close()

  return temp_file_path



import typing
from functools import partial, wraps
from types import GeneratorType, UnionType
from typing import Any
from typing import get_args as unpack_union
from warnings import warn

from _collections_abc import (
  Callable,
  Iterable,
  Mapping,
  bytearray_iterator,
  bytes_iterator,
  coroutine,
  dict_itemiterator,
  dict_items,
  dict_keyiterator,
  dict_keys,
  dict_valueiterator,
  dict_values,
  list_iterator,
  list_reverseiterator,
  longrange_iterator,
  range_iterator,
  set_iterator,
  str_iterator,
  tuple_iterator,
  zip_iterator,
)
from colored import Back, Fore, Style
from pydantic import BaseModel as PydanticBM

from .tcr_compare import able
from .tcr_constants import NEWLINE
from .tcr_extract_error import extract_error
from .tcr_int import hex as tcrhex
from .tcr_iterable import Or, getattr_queue, limited_iterable
from .tcr_null import Null, Undefined
from .tcr_types import QuotelessString


def print_block(
  text: str,
  border_char: str = '#',
  *,
  margin: int = 1,
  border: int = 3,
  padding: int = 0,
  padding_top: int = 1,
  padding_bottom: int = 1,
  text_color: str = Fore.YELLOW + Style.bold,
  border_color: str = Fore.WHITE + Style.bold,
  raw: bool = False,
  allow_invalid_config: bool = False,
) -> str | None:
  """Print or return a string of a "comment-like" block with text. Colors may optionally be set to `''` (empty string) to skip coloring of that part. Ends with a color reset unless none of the colors are enabled (both set to `''`).

  Params:
      - `margin`: The amount of spaces between the text and the inner walls (left-right only)
      - `border`: The width of walls (number of border_char characters, left-right only)
      - `padding`: The number of extra spaces added to the left of each line (left side only)
      - `padding_top` & `padding_bottom`: How many '\\n' characters to add at the beginning and the end of the string (top-bottom only)
  """

  text = str(text)

  if not allow_invalid_config and margin < 0 or border < 0 or padding < 0:
    msg = f'Invalid margin, border and/or padding(s) configuration {(margin, border, padding, padding_top, padding_bottom)!r}. Override this by passing in allow_invalid_config=True'
    raise ValueError(msg)

  if not allow_invalid_config and len(border_char) != 1:
    msg = f'border_char must be 1 character long (got {border_char!r} which is {len(border_char)!r} characters long). Override this by passing in allow_invalid_config=True'
    raise ValueError(msg)

  reset = Style.reset

  if not text_color and not border_color:
    reset = ''

  bar = f'{border_char * (border + margin + len(text) + margin + border)}'
  block = f"""
{padding_top * NEWLINE}{padding * ' '}{reset}{border_color}{bar}
{padding * ' '}{border * border_char}{reset}{margin * ' '}{text_color}{text}{reset}{margin * ' '}{border_color}{border * border_char}
{padding * ' '}{bar}{reset}{padding_bottom * NEWLINE}
"""[1:-1]
  if raw:
    return block

  print(block)
  return None


if True:  # \/ # fmt & print iterable
  # fmt: off
  # Format Colors, name kept short so lines don't get THAT long if this thing is used like 10 times in a single fstring
  class FMTC:
    _                   = Style.reset # Reset
    NUMBER              = Fore.BLUE + Style.bold
    TYPE                = Fore.BLUE + Style.bold
    DECIMAL             = Fore.WHITE + Style.bold
    BRACKET             = Fore.CYAN + Style.bold
    STRING              = Style.reset + Fore.YELLOW
    QUOTES              = Fore.WHITE + Style.bold
    COLON               = Fore.orange_1 + Style.bold
    ASTERISK            = Fore.orange_1 + Style.bold
    COROUTINE           = Fore.orange_1 + Style.bold
    FUNCTION            = Fore.orange_1 + Style.bold
    COMPLEX             = Fore.orange_1 + Style.bold
    COMMA               = Fore.dark_gray + Style.bold
    PIPE                = Fore.dark_gray + Style.bold
    UNKNOWN             = Fore.dark_gray + Style.bold
    TRUE                = Fore.GREEN + Style.bold
    FALSE               = Fore.RED + Style.bold
    NULL                = Fore.dark_gray + Style.bold
    UNDEFINED           = Fore.dark_gray + Style.bold
    NONE                = Fore.light_gray + Style.bold
    BYTESTR_B           = Fore.RED + Style.bold
    ITER_I              = Fore.red_3b + Style.bold
    SPECIAL             = Fore.purple_1b + Style.bold
    INTERNAL_EXCEPTION  = Fore.red_3b + Style.bold
    BUILT_IN_EXCEPTION  = Fore.BLUE + Style.bold
    CLASS               = Fore.BLUE + Style.bold

  class FMT_LETTERS:
    b  = f'{FMTC.BYTESTR_B}b'
    i  = f'{FMTC.ITER_I}i{FMTC._}'
    F  = f'{FMTC.SPECIAL}F'
    K  = f'{FMTC.SPECIAL}K'
    V  = f'{FMTC.SPECIAL}V'
    I  = f'{FMTC.SPECIAL}I'
    C  = f'{FMTC.COROUTINE}C'

  # Format Brackets templates.
  # (FMT_BRACKETS[_t][syntax_highlighting: bool] % content) -> attaches brackets to the content with respect to syntax highlighting
  FMT_BRACKETS: dict[type | None, tuple[str, str]] = { # Format Brackets
    None:          ('[%s]',            f'{FMTC.BRACKET}[{FMTC._}%s{FMTC.BRACKET}]{FMTC._}'),
    list:          ('[%s]',            f'{FMTC.BRACKET}[{FMTC._}%s{FMTC.BRACKET}]{FMTC._}'),
    dict_keys:     ('[%s]',            f'{FMT_LETTERS.K}{FMTC.BRACKET}[{FMTC._}%s{FMTC.BRACKET}]{FMTC._}'),
    dict_values:   ('[%s]',            f'{FMT_LETTERS.V}{FMTC.BRACKET}[{FMTC._}%s{FMTC.BRACKET}]{FMTC._}'),
    dict_items:    ('[%s]',            f'{FMT_LETTERS.I}{FMTC.BRACKET}[{FMTC._}%s{FMTC.BRACKET}]{FMTC._}'),
    set:           ('{%s}',            f'{FMTC.BRACKET}{{{FMTC._}%s{FMTC.BRACKET}}}{FMTC._}'),
    frozenset:     ('frozenset({%s})', f'{FMT_LETTERS.F}{FMTC.BRACKET}{{{FMTC._}%s{FMTC.BRACKET}}}{FMTC._}'),
    dict:          ('{%s}',            f'{FMTC.BRACKET}{{{FMTC._}%s{FMTC.BRACKET}}}{FMTC._}'),
    Mapping:       ('{%s}',            f'{FMTC.BRACKET}{{{FMTC._}%s{FMTC.BRACKET}}}{FMTC._}'),
    tuple:         ('(%s)',            f'{FMTC.BRACKET}({FMTC._}%s{FMTC.BRACKET}){FMTC._}'),
    GeneratorType: ('[%s]',            f'{FMTC.BRACKET}<{FMTC._}%s{FMTC.BRACKET}>{FMTC._}'),
    range:         ('range([%s])',     f'{FMTC.BRACKET}<{FMTC._}%s{FMTC.BRACKET}>{FMTC._}'),
    bytearray:     ('bytearray([%s])', f'{FMT_LETTERS.b}{FMTC.BRACKET}[{FMTC._}%s{FMTC.BRACKET}]{FMTC._}'),
    coroutine:     ('coroutine_%s',    f'{FMT_LETTERS.C}{FMTC.DECIMAL}\'{FMTC.FUNCTION}%s')
  }

  # FMT_TOO_DEEP[syntax_highlighting: bool] -> attaches '[ ... ]' to the content with respect to syntax highlighting
  FMT_TOO_DEEP = ('[ ... ]', f'{FMTC.BRACKET}[ {FMTC.DECIMAL}... {FMTC.BRACKET}]{FMTC._}')
  # FMT_UNION_SEPARATOR[syntax_highlighting: bool] -> attaches ' | ' to the content with respect to syntax highlighting
  FMT_UNION_SEPARATOR = ('|', f'{FMTC.PIPE} | {FMTC._}')
  # (FMT_ITER[syntax_highlighting: bool] % content) -> attaches 'i' or 'iter()' to the content with respect to syntax highlighting
  FMT_ITER = ('iter(%s)', f'{FMT_LETTERS.i}%s')
  # (FMT_UNKNOWN[syntax_highlighting: bool] % (name, content)) -> attaches name and content to an unknown object
  FMT_UNKNOWN = ('%s(%s)', f'{FMTC.UNKNOWN}%s({FMTC._}%s{FMTC.UNKNOWN}){FMTC._}')
  # (FMT_EXCEPTION[syntax_highlighting: bool] % getattr_queue(obj, '__name__', '__class__.__name__') -> Self explainatory
  FMT_INTERNAL_EXCEPTION = ("An exception occured while trying to display this item (%s).", f"{FMTC.INTERNAL_EXCEPTION}An exception occured while trying to display this item ({FMTC.UNKNOWN}%s{FMTC.INTERNAL_EXCEPTION}).{FMTC._}")
  # (FMT_CLASS[syntax_highlighting: bool]) % (getattr_queue(obj, '__name__', '__class__.__name__'), this(it))
  FMT_CLASS = ("%s(%s)", f"{FMTC.CLASS}%s{FMTC.BRACKET}({FMTC._}%s{FMTC.BRACKET}){FMTC._}")

  FMT_ASTERISK = ('*', f'{FMTC.ASTERISK}*')

  # fmt: on

  class _OverflowClass:
    amount: int

    def __init__(self, amount) -> None:
      self.amount = amount

    def __str__(self) -> str:
      return str(self.amount) if self.amount != -1 else '?'

  def _reset_return_wrapper(func):
    @wraps(func)
    def wrapper(*args, append_syntax_reset: bool = True, **kwargs):
      if kwargs.get('syntax_highlighting') and append_syntax_reset:
        return f'{FMTC._}{func(*args, **kwargs)}{FMTC._}'
      return func(*args, **kwargs)

    return wrapper

  def _non_double_repr(obj: object, /) -> str:
    rep = repr(obj)

    if not rep:
      return rep

    if not hasattr(obj, '__class__'):
      return rep

    if not hasattr(obj.__class__, '__name__'):
      return rep

    if rep.removeprefix(obj.__class__.__name__) == rep:
      return rep

    if not rep.removeprefix(obj.__class__.__name__):
      return ''

    if rep.removeprefix(obj.__class__.__name__)[0] != '(':
      return rep

    if rep[-1] != ')':
      return rep

    return rep[(len(obj.__class__.__name__) + 1) : -1]

  def fmt_iterable(
    it: Iterable | Any,
    /,
    *its: Iterable | Any,
    indentation: int = 2,
    item_limit: int = 100,
    depth_limit: int = 100,
    syntax_highlighting: bool = False,
    trailing_commas: bool = True,
    int_formatter: Callable[[int], str] | None = None,
    let_no_indent: bool = True,
    force_no_indent: bool = False,
    force_no_spaces: bool = False,
    force_complex_parenthesis: bool = False,
    force_union_parenthesis: bool = True,
    prefer_full_names: bool = False,
    **kwargs,
  ) -> str:
    """### Return iterable as formatted string with optional syntax highlighting.

    This also supports primitive types to be included in the iterables although passing primitive types themselves is fine.

    For printing the iterable use tcr.print_iterable().\\
    For debugging use tcr.console().

    Example:
    ```py
    >>> fmt_iterable([10, 20, 30])
    '[\\n  10,\\n  20,\\n  30,\\n]'
    >>> fmt_iterable({10: 20, 30: 40})
    '{\\n  10: 20,\\n  30: 40,\\n}'
    ```

    Args:
        it: Iterable, The iterable to format
        *its: Iterable, The same as above, if provided, it = (it, *its), can be omitted.
        indentation: int, How many spaces between where the list bracket is placed and the list items.
        item_limit: int, How many items in a list can be displayed before it's cut off. Any values < 1 will be set back to 1
        syntax_highlighting: bool, Whether or not terminal color codes should be added to make the code/iterables perttier. This also slightly changes the syntax. For example `'{,}'` with syntax highlighting and `'set()'` without. If you wish to copy & paste the iterable from terminal back to code set this to False although it may work with it set to true as well. The extra syntax safety does not support generators and `"(X more items...)"` messages that come from `item_limit`.
        trailing_commas: bool, whether or not to add trailing commas to lists (`'[\\n  10,\\n  20,\\n]'` vs `'[\\n  10,\\n  20\\n]'`). Psst: this can be set to 2 to force trailing commas although it's not recommended to use this as a feature.
        int_formatter: Callable[[int], str], A function that formats integers to a string. For example tcr.hex will format all integers as hex values. The tcr.hex formatter is automatically used unless other one was supplied
        let_no_indent: bool, Let the function automatically determine good spots to remove excessive enter-indent-syntax (for example "[\\n  10,\\n  20\\n]" - The list has only two, non iterable items or when the list has only one iterable or non iterable item)
        force_no_indent: bool, Force the condition above (let_no_spaces), skip any conditions listed there, never add extra enters or indents even if it may make the result unreadable. This effectively makes `indent` or `let_no_indent` irrelevant. This also sets `trailing_commas` to False.
        force_no_spaces: bool, Force remove any extra spaces (not including the objects' values for example strings will still be displayed with spaces). It removes spaces for example after commas, colons, etc. This effectively enables `force_no_indent` thus making `indent` or `let_no_indent` irrelevant.
        force_complex_parenthesis: bool, Force parenthesis when displaying `complex` type for example `(3 + 1j)` instead of `3+1j`. This has no effect when syntax highlighting is turned off.
        prefer_full_names: bool, whether or not to use the full names of objects if possible.
        let_no_inder_max_iterables: int = 1, (advanced) override the limit for iterables for the let_no_indent feature
        let_no_inder_max_non_iterables: int = 4, (advanced) override the limit for non-iterables for the let_no_indent feature

    If no formatting is set for an object, it can be defined using the __tcr_display__ method.

    ```py
    class PrintableObj:
      def __tcr_display__(self=None, **kwargs) -> str:
        return 'tcr.fmt-able object' + ('\'s instance' if self is not None else '')

    >>> fmt_iterable([PrintableObj])
    'tcr.fmt-able object'
    >>> fmt_iterable([PrintableObj()])
    "tcr.fmt-able object's instance"
    ```
    """
    if depth_limit < 0:
      depth_limit = 0
    if kwargs.get('__depth', 0) > depth_limit:
      return FMT_TOO_DEEP[syntax_highlighting]

    if its:
      it = (it, *its)

    item_limit = int(item_limit)
    if item_limit < 0:
      item_limit = 1

    if (
      # if this feature is turned on:
      let_no_indent
      and not force_no_spaces
      # for all Iterables...
      and isinstance(it, Iterable)
      # ...that don't have already hardcoded displays
      and not isinstance(it, str | bytes)
      # If their length can be checked (e.g. not generators)
      and able(len, it)
      # And they have any length:
      and len(it) > 0
    ):
      if isinstance(it, Mapping):
        if len(it) == 1 and ((type(next(iter(it.values()))) in (int, float, complex, bool)) or next(iter(it.values())) in (None, Null, '', [], (), {}, set())):
          force_no_indent = -1
      else:
        # Case 1: If the iterable in question contains iterables
        # If there is at most 1 iterable in the outer iterable of iterables
        if len(it) <= Or(kwargs.get('let_no_inder_max_iterables'), 1) and any(isinstance(x, Iterable) for x in it):
          force_no_indent = -1
        # Case 2: If the outer iterable consists of non-iterables: If there are at most 4 non-iterables
        if all((not isinstance(x, Iterable)) or isinstance(x, str | bytes) or (able(len, x) and len(x) == 0) for x in it) and len(it) <= Or(kwargs.get('let_no_inder_max_non_iterables'), 4):
          force_no_indent = -1

    name = '__qualname__' if prefer_full_names else '__name__'
    space = ' ' if not force_no_spaces else ''
    indent = space * indentation if not force_no_indent else ''
    enter = '\n' if not force_no_indent else ''
    trailing_commas = trailing_commas if (trailing_commas == 2 or not force_no_indent) else False

    if isinstance(it, Mapping | PydanticBM):
      asterisks = FMT_ASTERISK[syntax_highlighting] * 2
    elif isinstance(it, Iterable):
      asterisks = FMT_ASTERISK[syntax_highlighting] * 1
    else:
      asterisks = ''

    if int_formatter is None and isinstance(it, bytearray):
      int_formatter = tcrhex

    if force_no_indent < 0:
      force_no_indent += 1

    thisdict = {
      'indentation': indentation,
      'item_limit': item_limit,
      'syntax_highlighting': syntax_highlighting,
      'trailing_commas': trailing_commas,
      'int_formatter': int_formatter,
      'let_no_indent': let_no_indent,
      'force_no_indent': force_no_indent,
      'force_no_spaces': force_no_spaces,
      'force_complex_parenthesis': force_complex_parenthesis,
      'force_union_parenthesis': force_union_parenthesis,
      'prefer_full_names': prefer_full_names,
      'depth_limit': depth_limit,
      '__depth': kwargs.get('__depth', 2) + 1,
    }
    if a := kwargs.get('let_no_inder_max_iterables'):
      thisdict['let_no_inder_max_iterables'] = a
    if a := kwargs.get('let_no_inder_max_non_iterables'):
      thisdict['let_no_inder_max_non_iterables'] = a

    this = partial(fmt_iterable, **thisdict)

    if '_force_next_type' in kwargs and kwargs.get('_ran_from_tcr_display'):
      queue_name = getattr_queue(
        kwargs['_force_next_type'],
        name,
        '__name__',
        '__class__.__name__',
        default=('<???>' if syntax_highlighting else '__unknown_object__'),
      )

      return (FMT_LETTERS.C if syntax_highlighting and kwargs.get('_i_am_class') else '') + (
        FMT_CLASS[syntax_highlighting]
        % (
          queue_name + ('(...)' if not syntax_highlighting and not kwargs.get('_i_am_class') else ''),
          asterisks + this(it),
        )
      )

    if isinstance(it, _OverflowClass):
      return f'{FMTC.SPECIAL}({FMTC.NUMBER}{it}{FMTC.SPECIAL} more item{"s" if it.amount != 1 else ""}...){FMTC._}' if syntax_highlighting else f'({it} more items...)'
    if isinstance(it, PydanticBM):
      return FMT_CLASS[syntax_highlighting] % (it.__class__.__name__, asterisks + this(it.model_dump()))

    if able(issubclass, it, BaseException) and issubclass(it, BaseException):
      exc_name = extract_error(it, raw=True)[0]

      if not syntax_highlighting:
        return exc_name

      return f'{FMTC.BUILT_IN_EXCEPTION}{exc_name}'

    queue_name = getattr_queue(
      type(it),
      name,
      '__name__',
      '__class__.__name__',
      default=('<???>' if syntax_highlighting else '__unknown_object__'),
    )

    if hasattr(it, '__tcr_display__'):
      try:
        return it.__tcr_display__(**thisdict, _ran_from_tcr_display=True)
      except NotImplementedError:
        pass
      except Exception as e:
        if kwargs.get('_raise_errors'):
          raise
        return FMT_INTERNAL_EXCEPTION[syntax_highlighting] % f'{queue_name}, {extract_error(e, raw=True)[0]}'

    if hasattr(it, '__tcr_fmt__'):
      try:
        return it.__tcr_fmt__(**thisdict, fmt_iterable=fmt_iterable, _ran_from_tcr_display=True)
      except NotImplementedError:
        pass
      except Exception as e:
        if kwargs.get('_raise_errors'):
          raise
        return FMT_INTERNAL_EXCEPTION[syntax_highlighting] % f'{queue_name}, {extract_error(e, raw=True)[0]}'

    if not syntax_highlighting and isinstance(it, QuotelessString):
      return str(it)

    if it is Null:
      return f'{FMTC.NULL}{it}{FMTC._}' if syntax_highlighting else str(it)
    if it is Undefined:
      return f'{FMTC.UNDEFINED}{it}{FMTC._}' if syntax_highlighting else str(it)
    if it is None:
      return f'{FMTC.NONE}{it}{FMTC._}' if syntax_highlighting else str(it)
    if it is True:
      return f'{FMTC.TRUE}{it}{FMTC._}' if syntax_highlighting else str(it)
    if it is False:
      return f'{FMTC.FALSE}{it}{FMTC._}' if syntax_highlighting else str(it)

    _t = kwargs.get('_force_next_type') or type(it)

    if _t == QuotelessString:
      reprit = repr(it)
      return f'{FMTC.STRING}{reprit[1:-1]}' if syntax_highlighting else repr(it)

    if _t == type:
      return f'{FMTC.TYPE}{getattr_queue(it, "__name__", "__class__.__name__", default=it)}{FMTC._}' if syntax_highlighting else str(it)

    if not kwargs.get('_force_next_type'):
      itsmro = getattr_queue(it, '__mro__', '__class__.__mro__', default=())
      # It's-a me mro

      if str in itsmro:
        _t = str  # Things displayed as ['a', 'b', 'c', 'd'] if they inherited from str
      elif bytes in itsmro:
        _t = bytes
      else:
        for item in itsmro:
          if item in FMT_BRACKETS:
            _t = item
            break
    if _t == int:
      if int_formatter:
        it = int_formatter(it)
      return f'{FMTC.NUMBER}{it}{FMTC._}' if syntax_highlighting else str(it)
    if _t == float:
      return f'{FMTC.NUMBER}{str(it).replace(".", f"{FMTC.DECIMAL}.{FMTC.NUMBER}")}{FMTC._}' if syntax_highlighting else str(it)
    if _t == str:
      reprit = repr(it)
      return f'{FMTC.QUOTES}{reprit[0]}{FMTC.STRING}{reprit[1:-1]}{FMTC.QUOTES}{reprit[-1]}{FMTC._}' if syntax_highlighting else repr(it)
    if _t == bytes:
      reprit = repr(it)
      return f'{FMTC.BYTESTR_B}{reprit[0]}{FMTC.QUOTES}{reprit[1]}{FMTC.STRING}{reprit[2:-1]}{FMTC.QUOTES}{reprit[-1]}{FMTC._}' if syntax_highlighting else repr(it)
    if _t == complex:
      brackets = ('', '') if (not force_complex_parenthesis) else (f'{FMTC.BRACKET}(', f'{FMTC.BRACKET})')
      return (
        f"""{brackets[0]}{FMTC.NUMBER}{int(it.real) if int(it.real) == it.real else str(it.real).replace(".", f"{FMTC.DECIMAL}.{FMTC.NUMBER}")}{FMTC._}{space}{FMTC.COMPLEX}+{space}{FMTC.NUMBER}{int(it.imag) if int(it.imag) == it.imag else str(it.imag).replace(".", f"{FMTC.DECIMAL}.{FMTC.NUMBER}")}{FMTC.COMPLEX}j{brackets[1]}{FMTC._}"""
        if syntax_highlighting
        else repr(it)
      )
    if _t == UnionType:
      if force_union_parenthesis:
        return FMT_BRACKETS[tuple][syntax_highlighting] % (FMT_UNION_SEPARATOR[syntax_highlighting].join(this(x) for x in unpack_union(it)))
      else:
        return FMT_UNION_SEPARATOR[syntax_highlighting].join(this(x) for x in unpack_union(it))

    if _t == bytes_iterator:
      return FMT_ITER[syntax_highlighting] % this(bytes(it))
    if _t == str_iterator:
      return FMT_ITER[syntax_highlighting] % this(''.join(it))
    if _t == bytearray_iterator:
      return FMT_ITER[syntax_highlighting] % this(bytearray(it))
    if _t == dict_keyiterator:
      listit, ov = limited_iterable(it, item_limit)
      return FMT_ITER[syntax_highlighting] % this(listit, _force_next_type=dict_keys, _ov=ov)
    if _t == dict_valueiterator:
      listit, ov = limited_iterable(it, item_limit)
      return FMT_ITER[syntax_highlighting] % this(listit, _force_next_type=dict_values, _ov=ov)
    if _t == dict_itemiterator:
      listit, ov = limited_iterable(it, item_limit)
      return FMT_ITER[syntax_highlighting] % this(listit, _force_next_type=dict_items, _ov=ov)
    if _t in (list_iterator, list_reverseiterator, zip_iterator):
      listit, ov = limited_iterable(it, item_limit)
      return FMT_ITER[syntax_highlighting] % this(listit, _ov=ov)
    if _t in (range_iterator, longrange_iterator):
      listit, ov = limited_iterable(it, item_limit)
      return FMT_ITER[syntax_highlighting] % this(listit, _force_next_type=range, _ov=ov)
    if _t == set_iterator:
      listit, ov = limited_iterable(it, item_limit)
      return FMT_ITER[syntax_highlighting] % this(listit, _force_next_type=set, _ov=ov)
    if _t == tuple_iterator:
      listit, ov = limited_iterable(it, item_limit)
      return FMT_ITER[syntax_highlighting] % this(listit, _force_next_type=tuple, _ov=ov)
    if _t == coroutine:
      return FMT_BRACKETS[coroutine][syntax_highlighting] % (getattr_queue(it, name, '__name__'))

    comma = f'{FMTC.COMMA},' if syntax_highlighting else ','

    is_of_prohibited_type = any(isinstance(it, x) for x in (typing._GenericAlias, typing._UnpackGenericAlias))
    # took me half an hour of debugging to discover that typing._GenericAlias contains an infinite amount of nested copies of typing._UnpackGenericAlias but only if you first iter then index not if you just index and it's a mess.

    if not is_of_prohibited_type and isinstance(it, Iterable):
      itl, overflow = limited_iterable(it, item_limit)
      if not able(len, it) or len(it) > 0:
        if isinstance(it, Mapping):
          inner = f'{comma}{enter or space}'.join(
            [
              indent + (f'{k}{FMTC.COLON}:{FMTC._}{space}{v}' if syntax_highlighting else f'{k}:{space}{v}').replace('\n', f'{enter}{indent}')
              for k, v in {this(key): this(value) for key, value in it.items()}.items()
            ]
          ) + (comma if trailing_commas else '')

          return (FMT_BRACKETS[_t] if _t in FMT_BRACKETS else FMT_BRACKETS[dict])[syntax_highlighting] % f'{enter}{inner}{enter}'  # fmt: skip
        else:
          inner = f'{comma}{enter or space}'.join(
            [
              indent + x.replace('\n', f'\n{indent}')
              for x in [this(element) for element in itl]
              + ([] if not (overflow if not kwargs.get('_ov') else kwargs.get('_ov')) else [this(_OverflowClass(overflow if not kwargs.get('_ov') else kwargs.get('_ov')))])
            ]
          ) + (comma if (trailing_commas or (_t == tuple and len(it) == 1)) else '')

          return (FMT_BRACKETS[_t] if _t in FMT_BRACKETS else FMT_BRACKETS[None])[syntax_highlighting] % f'{enter}{inner}{enter}'  # fmt: skip
      else:
        if _t == set and not syntax_highlighting:
          return 'set()'
        return (FMT_BRACKETS[_t] if _t in FMT_BRACKETS else FMT_BRACKETS[None])[syntax_highlighting] % (comma if _t == set else '')  # fmt: skip

    return FMT_UNKNOWN[syntax_highlighting] % (
      queue_name,
      _non_double_repr(it),
    )  # If no hardcoded patterns match, return a repred version of whatever it is

  globals()['fmt_iterable'] = _reset_return_wrapper(fmt_iterable)
  # Done this way to trick the IDE code snippets since the deco forwards the *args, **kwargs to the decorated func anyway...
  # If i were to @decorate the def fmt_iterable it'd replace the code snippets with the decorated function's code snippets.
  # Or i'm a dumbass and there's a better way of doing it.. (<- probably this)

  def print_iterable(
    it: Iterable,
    *its: Iterable,
    raw: bool | None = None,
    recursive: bool | None = None,
    printhook: Callable[[str], None] = print,
    **kwargs,
  ) -> None:
    """### Print iterable as formatted string with optional syntax highlighting with the given printhook.

    This also supports primitive types to be included in the iterables although passing primitive types themselves is fine.

    For formatting the iterable (and having the result returned as str) use tcr.fmt_iterable().\\
    For debugging use tcr.console().

    Example:
    ```py
    >>> print_iterable([10, 20, 30]) # Returns None, if you need str use fmt_iterable()
    [
      10,
      20,
      30,
    ]
    >>> print_iterable({10: 20, 30: 40})
    {
      10: 20,
      30: 40,
    }
    ```

    Args:
        it: Iterable, The iterable to print
        *its: Iterable, The same as above, if provided, it = (it, *its), can be omitted.
        indentation: int, How many spaces between where the list bracket is placed and the list items.
        raw: DEPRECATED, don't use. If you need raw output use fmt_iterable()
        recursive: DEPRECATED, this does not have any effect but is left for backwards-compatibility.
        printhook: Callable[[str], None], replacement for print function if provided else the built in print function
        item_limit: int, How many items in a list can be displayed before it's cut off. Any values < 1 will be set back to 1
        syntax_highlighting: bool, Whether or not terminal color codes should be added to make the code/iterables perttier. This also slightly changes the syntax. For example `'{,}'` with syntax highlighting and `'set()'` without. If you wish to copy & paste the iterable from terminal back to code set this to False although it may work with it set to true as well. The extra syntax safety does not support generators and `"(X more items...)"` messages that come from `item_limit`.
        trailing_commas: bool, whether or not to add trailing commas to lists (`'[\\n  10,\\n  20,\\n]'` vs `'[\\n  10,\\n  20\\n]'`)
        int_formatter: Callable[[int], str]: A function that formats integers to a string. For example tcr.hex will format all integers as hex values. The tcr.hex formatter is automatically used unless other one was supplied
        let_no_indent: bool, Let the function automatically determine good spots to remove excessive enter-indent-syntax (for example "[\\n  10,\\n  20\\n]" - The list has only two, non iterable items or when the list has only one iterable or non iterable item)
        force_no_indent: bool, Force the condition above (let_no_spaces), skip any conditions listed there, never add extra enters or indents even if it may make the result unreadable. This effectively makes `indent` or `let_no_indent` irrelevant. This also sets `trailing_commas` to False.
        force_no_spaces: bool, Force remove any extra spaces (not including the objects' values for example strings will still be displayed with spaces). It removes spaces for example after commas, colons, etc. This effectively enables `force_no_indent` thus making `indent` or `let_no_indent` irrelevant.
        force_complex_parenthesis: bool, Force parenthesis when displaying `complex` type for example `(3 + 1j)` instead of `3+1j`. This has no effect when syntax highlighting is turned off.
        prefer_full_names: bool, whether or not to use the full names of objects if possible.
        **kwargs: Anything from there is passed in into the fmt_iterable call
    """
    result = fmt_iterable(it, *its, **kwargs)
    if recursive is not None:
      warn(
        "`recursive` is deprecated. It's always set to True in new print_iterable()",
        DeprecationWarning,
        stacklevel=2,
      )
    if raw:
      warn(
        '`raw` is deprecated. Use fmt_iterable() instead',
        DeprecationWarning,
        stacklevel=2,
      )
      return result
    printhook(result)


def alert(s: str, *, printhook: Callable[[str], None] = print, raw=False) -> None:
  text = ''.join([f'{(Fore.BLACK + Back.RED + Style.bold) if i % 2 == 0 else Fore.WHITE + Back.YELLOW}{x}' for i, x in enumerate(s)]) + Style.reset

  if raw:
    return text

  printhook(text)
